"""Routines and classes for representing periodic lines in R-Z
poloidal planes

"""

import itertools
import warnings

import numpy as np
from numpy import append, argmin, cos, linspace, pi, sin, sqrt
from scipy.integrate import cumulative_trapezoid as cumtrapz
from scipy.interpolate import interp1d, splev, splrep

try:
    import matplotlib.pyplot as plt

    plotting_available = True
except ImportError:
    warnings.warn("Couldn't import matplotlib, plotting not available.")
    plotting_available = False


class RZline:
    """Represents (R,Z) coordinates of a periodic line

    Attributes
    ----------
    R : array_like
        Major radius [m]
    Z : array_like
        Height [m]
    theta : array_like
        Angle variable [radians]

        `R`, `Z` and `theta` all have the same length

    Parameters
    ----------
    r, z : array_like
        1D arrays of the major radius (`r`) and height (`z`) which are
        of the same length. A periodic domain is assumed, so the last
        point connects to the first.

    anticlockwise : bool, optional
        Ensure that the line goes anticlockwise in the R-Z plane
        (positive theta)

    spline_order : int, optional
        Change the spline order for scipy.interpolate.splrep

    Note that the last point in (r,z) arrays should not be the same
    as the first point. The (r,z) points are in [0,2pi)

    The input r,z points will be reordered, so that the
    theta angle goes anticlockwise in the R-Z plane

    """

    def __init__(self, r, z, anticlockwise=True, spline_order=None, smooth=False):
        r = np.asarray(r)
        z = np.asarray(z)

        # Check the sizes of the variables
        n = len(r)
        assert len(z) == n
        assert r.shape == (n,)
        assert z.shape == (n,)

        if anticlockwise:
            # Ensure that the line is going anticlockwise (positive theta)
            # The first method is faster, but unreliable.
            # mid_ind = np.argmax(r)  # Outboard midplane index
            # ind_next = (mid_ind + 1) % n

            # if z[ind_next] < z[mid_ind]:
            #     # Line going down at outboard midplane. Need to reverse
            #     r = r[::-1]  # r = np.flip(r)
            #     z = z[::-1]  # z = np.flip(z)
            # Calcculating the area should be much more robust
            A = np.sum((r - np.roll(r, 1)) * (z + np.roll(z, 1)))
            assert A != 0
            if A > 0:
                r = r[::-1]  # r = np.flip(r)
                z = z[::-1]  # z = np.flip(z)
            assert np.sum((r - np.roll(r, 1)) * (z + np.roll(z, 1))) < 0

        self.R = r
        self.Z = z

        # Define an angle variable
        self.theta = linspace(0, 2 * pi, n, endpoint=False)

        self.spline_order = spline_order or 3

        # Create a spline representation
        # Note that the last point needs to be passed but is not used
        kw = dict(per=True, k=self.spline_order)
        if smooth:
            if smooth is True:
                smooth = 100
            num = len(r) // smooth + 1
            kw["t"] = np.linspace(0, np.pi * 2, num, endpoint=False) + np.pi / num
        self._rspl = splrep(append(self.theta, 2 * pi), append(r, r[0]), **kw)
        self._zspl = splrep(append(self.theta, 2 * pi), append(z, z[0]), **kw)

    def Rvalue(self, theta=None, deriv=0):
        """Calculate the value of R at given theta locations

        Parameters
        ----------
        theta : array_like, optional
            Theta locations to find R at. If None (default), use the
            values of theta stored in the instance
        deriv : int, optional
            The order of derivative to compute (default is just the R value)

        Returns
        -------
        ndarray
            Value of R at each input theta point
        """
        if theta is None:
            theta = self.theta
        else:
            theta = np.remainder(theta, 2 * np.pi)

        return splev(theta, self._rspl, der=deriv)

    def Zvalue(self, theta=None, deriv=0):
        """Calculate the value of Z at given theta locations

        Parameters
        ----------
        theta : array_like, optional
            Theta locations to find Z at. If None (default), use the
            values of theta stored in the instance
        deriv : int, optional
            The order of derivative to compute (default is just the Z value)

        Returns
        -------
        ndarray
            Value of Z at each input theta point
        """
        if theta is None:
            theta = self.theta
        else:
            theta = np.remainder(theta, 2 * np.pi)
        return splev(theta, self._zspl, der=deriv)

    def position(self, theta=None):
        """Calculate the value of both R, Z at given theta locations

        Parameters
        ----------
        theta : array_like, optional
            Theta locations to find R, Z at. If None (default), use the
            values of theta stored in the instance

        Returns
        -------
        R, Z : (ndarray, ndarray)
            Value of R, Z at each input theta point
        """
        return self.Rvalue(theta=theta), self.Zvalue(theta=theta)

    def positionPolygon(self, theta=None):
        """Calculates (R,Z) position at given theta angle by joining points
        by straight lines rather than a spline. This avoids the
        overshoots which can occur with splines.

        Parameters
        ----------
        theta : array_like, optional
            Theta locations to find R, Z at. If None (default), use the
            values of theta stored in the instance

        Returns
        -------
        R, Z : (ndarray, ndarray)
            Value of R, Z at each input theta point
        """
        if theta is None:
            return self.R, self.Z
        n = len(self.R)
        dtheta = 2.0 * np.pi / n
        ind = np.trunc(theta / dtheta).astype(int) % n
        rem = np.remainder(theta, dtheta)
        indp = (ind + 1) % n
        return (rem * self.R[indp] + (1.0 - rem) * self.R[ind]), (
            rem * self.Z[indp] + (1.0 - rem) * self.Z[ind]
        )

    def distance(self, sample=20, weights=None):
        """Integrates the distance along the line.

        Parameters
        ----------
        sample : int, optional
            Number of samples to take per point

        Returns
        -------
        An array one longer than theta. The first element is zero,
        and the last element is the total distance around the loop

        """

        def interp(dat, n):
            t1 = np.linspace(0, 1, len(dat) + 1)
            return interp1d(t1, np.append(dat, dat[0]), assume_sorted=True)(
                np.linspace(0, 1, n, endpoint=False)
            )

        if self.spline_order == 1:
            R = self.R
            Z = self.Z
            dr = (R - np.roll(R, -1)) ** 2 + (Z - np.roll(Z, -1)) ** 2
            dr = np.sqrt(dr)
            weights = (
                interp(weights, len(dr)) if not weights is None else itertools.repeat(1)
            )
            out = np.empty(len(dr) + 1)
            sum = 0
            for i, (c, w) in enumerate(zip(dr, weights)):
                out[i] = sum
                sum += c * w
            out[-1] = sum
            return out

        sample = int(sample)
        assert sample >= 1

        thetavals = np.linspace(
            0.0, 2.0 * np.pi, sample * len(self.theta) + 1, endpoint=True
        )
        # variation of length with angle dl/dtheta
        dldtheta = sqrt(
            self.Rvalue(thetavals, deriv=1) ** 2 + self.Zvalue(thetavals, deriv=1) ** 2
        )
        if weights is not None:
            dldtheta *= interp(weights, len(thetavals))

        # Integrate cumulatively, then take only the values at the grid points (including end)
        return cumtrapz(dldtheta, thetavals, initial=0.0)[::sample]

    def equallySpaced(self, n=None, weights=None):
        """Returns a new RZline which has a theta uniform in distance along
        the line

        Parameters
        ----------
        n : int, optional
            Number of points. Default is the same as the current line

        Returns
        -------
        `RZline`
            A new `RZline` based on this instance, but with uniform theta-spacing
        """
        if n is None:
            n = len(self.theta)

        # Distance along the line
        dist = self.distance(weights=weights)

        # Positions where points are desired
        positions = linspace(dist[0], dist[-1], n, endpoint=False)

        # Find which theta value these correspond to
        thetavals = interp1d(
            dist, append(self.theta, 2.0 * pi), copy=False, assume_sorted=True
        )
        new_theta = thetavals(positions)

        return RZline(self.Rvalue(new_theta), self.Zvalue(new_theta))

    def closestPoint(self, R, Z, niter=3, subdivide=20):
        """Find the closest point on the curve to the given (R,Z) point

        Parameters
        ----------
        R, Z : float
            The input R, Z point
        niter : int, optional
            How many iterations to use

        Returns
        -------
        float
            The value of theta (angle)

        """

        # First find the closest control point
        ind = argmin((self.R - R) ** 2 + (self.Z - Z) ** 2)
        theta0 = self.theta[ind]
        dtheta = self.theta[1] - self.theta[0]

        # Iteratively refine and find new minimum
        for i in range(niter):
            # Create a new set of points between point (ind +/- 1)
            # By using dtheta, wrapping around [0,2pi] is handled
            thetas = np.linspace(
                theta0 - dtheta, theta0 + dtheta, subdivide, endpoint=False
            )
            Rpos, Zpos = self.positionPolygon(thetas)

            ind = argmin((Rpos - R) ** 2 + (Zpos - Z) ** 2)
            theta0 = thetas[ind]
            dtheta = thetas[1] - thetas[0]

        return np.remainder(theta0, 2 * np.pi)

    def plot(self, axis=None, show=True):
        """Plot the RZline, either on the given axis or a new figure

        Parameters
        ----------
        axis : matplotlib axis, optional
            A matplotlib axis to plot on. By default a new figure
            is created
        show : bool, optional
            Calls plt.show() at the end

        Returns
        -------
        axis
            The matplotlib axis that was used

        """

        if not plotting_available:
            warnings.warn("matplotlib not available, unable to plot")
            return None

        if axis is None:
            fig = plt.figure()
            axis = fig.add_subplot(1, 1, 1)

        theta = np.linspace(0, 2 * np.pi, 100, endpoint=True)
        axis.plot(self.Rvalue(theta), self.Zvalue(theta), "k-")
        axis.plot(self.R, self.Z, "ro")

        if show:
            plt.show()

        return axis


def circle(R0=1.0, r=0.5, n=20):
    """Creates a pair of RZline objects, for inner and outer boundaries

    Parameters
    ----------
    R0 : float, optional
        Centre point of the circle
    r : float, optional
        Radius of the circle
    n : int, optional
        Number of points to use in the boundary

    Returns
    -------
    `RZline`
        A circular `RZline`

    """
    # Define an angle coordinate
    theta = linspace(0, 2 * pi, n, endpoint=False)

    return RZline(R0 + r * cos(theta), r * sin(theta))


def shaped_line(R0=3.0, a=1.0, elong=0.0, triang=0.0, indent=0.0, n=20):
    """Parametrisation of plasma shape from J. Manickam, Nucl. Fusion 24
    595 (1984)

    Parameters
    ----------
    R0 : float, optional
        Major radius
    a : float, optional
        Minor radius
    elong : float, optional
        Elongation, 0 for a circle
    triang : float, optional
        Triangularity, 0 for a circle
    indent : float, optional
        Indentation, 0 for a circle

    Returns
    -------
    `RZline`
        An `RZline` matching the given parameterisation

    """
    theta = linspace(0, 2 * pi, n, endpoint=False)
    return RZline(
        R0 - indent + (a + indent * cos(theta)) * cos(theta + triang * sin(theta)),
        (1.0 + elong) * a * sin(theta),
    )


def line_from_points_poly(rarray, zarray, show=False, spline_order=None):
    """Find a periodic line which goes through the given (r,z) points

    This function starts with a triangle, then adds points
    one by one, inserting into the polygon along the nearest
    edge

    Parameters
    ----------
    rarray, zarray : array_like
        R, Z coordinates. These arrays should be the same length

    Returns
    -------
    `RZline`
        An `RZline` object representing a periodic line

    """

    rarray = np.asarray(rarray)
    zarray = np.asarray(zarray)

    assert rarray.size >= 3
    assert rarray.shape == zarray.shape

    npoints = rarray.size

    rvals = rarray.copy()
    zvals = zarray.copy()

    # Take the first three points to make a triangle

    if show and plotting_available:
        plt.figure()
        plt.plot(rarray, zarray, "x")
        plt.plot(
            np.append(rvals[:3], rvals[0]), np.append(zvals[:3], zvals[0])
        )  # Starting triangle

    for i in range(3, npoints):
        line = RZline(rvals[:i], zvals[:i], spline_order=spline_order)

        angle = np.linspace(0, 2 * pi, 100)
        r, z = line.position(angle)

        # Next point to add
        # plt.plot(rarray[i], zarray[i], 'o')

        # Find the closest point on the line
        theta = line.closestPoint(rarray[i], zarray[i])

        rl, zl = line.position(theta)

        ind = int(np.floor(float(i) * theta / (2.0 * np.pi)))

        # Insert after this index

        if ind != i - 1:
            # If not the last point, then need to shift other points along
            rvals[ind + 2 : i + 1] = rvals[ind + 1 : i]
            zvals[ind + 2 : i + 1] = zvals[ind + 1 : i]
        rvals[ind + 1] = rarray[i]
        zvals[ind + 1] = zarray[i]

        if show and plotting_available:
            plt.plot([rarray[i], rl], [zarray[i], zl])
            plt.plot(
                np.append(rvals[: (i + 1)], rvals[0]),
                np.append(zvals[: (i + 1)], zvals[0]),
            )  # New line

    if show and plotting_available:
        plt.show()
    return RZline(rvals, zvals, spline_order=spline_order)


def line_from_points_fast(rarray, zarray, **kw):
    """
    Find a periodic line which goes through the given (r,z) points.

    This version is particularly fast, but fails if the points are not
    sufficiently close to a circle.
    """
    rz = rarray, zarray
    cent = [np.mean(x) for x in rz]
    dist = [a - b for (a, b) in zip(rz, cent)]
    angle = np.arctan2(*dist)
    ind = np.argsort(angle)
    return RZline(*[x[ind] for x in rz], **kw)


def line_from_points_two_opt(rarray, zarray, opt=1e-3, **kwargs):
    """This is probably way to slow.

    Use the two opt algorithm:
    From https://stackoverflow.com/a/44080908
    License: CC-BY-SA 4.0
    """
    # Calculate the euclidian distance in n-space of the route r traversing cities c, ending at the path start.
    path_distance = lambda r, c: np.sum(
        [np.linalg.norm(c[r[p]] - c[r[p - 1]]) for p in range(len(r))]
    )
    # Reverse the order of all elements from element i to element k in array r.
    two_opt_swap = lambda r, i, k: np.concatenate(
        (r[0:i], r[k : -len(r) + i - 1 : -1], r[k + 1 : len(r)])
    )

    # 2-opt Algorithm adapted from https://en.wikipedia.org/wiki/2-opt
    def two_opt(cities, improvement_threshold):
        # Make an array of row numbers corresponding to cities.
        route = np.arange(cities.shape[0])
        improvement_factor = 1
        best_distance = path_distance(route, cities)
        while improvement_factor > improvement_threshold:
            # Record the distance at the beginning of the loop.
            distance_to_beat = best_distance
            # From each city except the first and last,
            for swap_first in range(1, len(route) - 2):
                # to each of the cities following,
                for swap_last in range(swap_first + 1, len(route)):
                    # try reversing the order of these cities
                    new_route = two_opt_swap(route, swap_first, swap_last)
                    # and check the total distance with this modification.
                    new_distance = path_distance(new_route, cities)
                    # If the path distance is an improvement,
                    if new_distance < best_distance:
                        route = new_route
                        best_distance = new_distance
                # Calculate how much the route has improved.
                improvement_factor = 1 - best_distance / distance_to_beat
        return route

    l = line_from_points_fast(rarray, zarray, **kwargs)
    cities = np.array([l.R, l.Z]).T
    route = two_opt(cities, opt)
    R, Z = cities[route].T
    return RZline(R, Z, **kwargs)


def line_from_points_convex_hull(rarray, zarray, **kwargs):
    """
    Find a periodic line which goes through the given (r,z) points

    This function uses an alphashape to find the boundary, and then projects
    the points on the boundary for sorting.

    This requires shapely
    """
    import shapely as sg

    rz = np.array((rarray, zarray)).T
    shape = sg.MultiPoint(rz).convex_hull.exterior
    theta = [shape.project(sg.Point(*p)) for p in rz]
    ind = np.argsort(theta)
    return RZline(*rz[ind].T)


def line_from_points_alphashape(rarray, zarray, alpha=2.0, **kwargs):
    """
    Find a periodic line which goes through the given (r,z) points

    This function uses an alphashape to find the boundary, and then projects
    the points on the boundary for sorting.

    This requires shapely and alphashape
    """
    import alphashape
    import shapely as sg

    rz = np.array((rarray, zarray)).T
    shape = alphashape.alphashape(rz, alpha)
    shape = shape.exterior
    theta = [shape.project(sg.Point(*p)) for p in rz]
    ind = np.argsort(theta)
    return RZline(*rz[ind].T)


def line_from_points(
    rarray, zarray, show=False, spline_order=None, is_sorted=False, smooth=False
):
    """Find a periodic line which goes through the given (r,z) points

    This function starts at a point, and finds the nearest neighbour
    which is not already in the line

    Parameters
    ----------
    rarray, zarray : array_like
        R, Z coordinates. These arrays should be the same length

    show : bool (optional)
        Whether to plot the found solution (using matplotlib)

    spline_order : integer (optional)
        Allows to change the spline order

    is_sorted : bool (optional)
        Avoids sorting the data. Makes it faster, but no checking is
        performed to ensure that the data is truly sorted.

    Returns
    -------
    `RZline`
        An `RZline` object representing a periodic line

    """

    # Make sure we have Numpy arrays
    rarray = np.asarray(rarray)
    zarray = np.asarray(zarray)

    assert rarray.size == zarray.size

    # We can get different answers depending on which point
    # we start the line on.
    # Therefore start the line from every point in turn,
    # and keep the line with the shortest total distance

    if is_sorted:
        return RZline(rarray, zarray, spline_order=spline_order, smooth=smooth)

    best_line = None  # The best line found so far
    best_dist = 0.0  # Distance around best line

    for start_ind in range(rarray.size):
        # Create an array of remaining points
        # Make copies since we edit the array later
        rarr = np.roll(rarray, start_ind).copy()
        zarr = np.roll(zarray, start_ind).copy()

        # Create new lists for the result
        rvals = [rarr[0]]
        zvals = [zarr[0]]

        rarr = rarr[1:]
        zarr = zarr[1:]

        while rarr.size > 1:
            # Find the index in array closest to last point
            ind = np.argmin((rvals[-1] - rarr) ** 2 + (zvals[-1] - zarr) ** 2)

            rvals.append(rarr[ind])
            zvals.append(zarr[ind])
            # Shift arrays
            rarr[ind:-1] = rarr[(ind + 1) :]
            zarr[ind:-1] = zarr[(ind + 1) :]
            # Chop off last point
            rarr = rarr[:-1]
            zarr = zarr[:-1]

        # One left, add to the end
        rvals.append(rarr[0])
        zvals.append(zarr[0])

        new_line = RZline(rvals, zvals, spline_order=spline_order, smooth=smooth)
        new_dist = new_line.distance()[-1]  # Total distance

        if (best_line is None) or (new_dist < best_dist):
            # Either if we haven't got a line, or found
            # a better line
            best_line = new_line
            best_dist = new_dist

    return best_line


if __name__ == "__main__":
    import field
    import fieldtracer
    import poloidal_grid

    #############################################################################
    # Define the magnetic field
    # Length in y after which the coils return to their starting (R,Z) locations
    yperiod = 10.0

    magnetic_field = field.StraightStellarator(I_coil=0.3, radius=1.0, yperiod=yperiod)

    #############################################################################
    # Create the inner flux surface, starting at a point at phi=0
    # To do this we need to define the y locations of the poloidal points
    # where we will construct grids

    start_r = 0.2
    start_z = 0.0

    nslices = 8  # Number of poloidal slices
    ycoords = np.linspace(0, yperiod, nslices)
    npoints = 20  # Points per poloidal slice

    # Create a field line tracer
    tracer = fieldtracer.FieldTracer(magnetic_field)

    # Extend the y coordinates so the tracer loops npoints times around yperiod
    ycoords_all = ycoords
    for i in range(1, npoints):
        ycoords_all = np.append(ycoords_all, ycoords + i * yperiod)

    coord = tracer.follow_field_lines(start_r, start_z, ycoords_all, rtol=1e-12)

    inner_lines = []
    for i in range(nslices):
        r = coord[i::nslices, 0]
        z = coord[i::nslices, 1]
        line = line_from_points(r, z)
        # Re-map the points so they're approximately uniform in distance along the surface
        # Note that this results in some motion of the line
        line = line.equallySpaced()
        inner_lines.append(line)

    # Now have a list of y coordinates (ycoords) and inner lines (inner_lines)

    #############################################################################
    # Generate a fixed circle for the outer boundary

    outer_line = circle(R0=0.0, r=0.8)

    #############################################################################
    # Now have inner and outer boundaries for each poloidal slice
    # Generate a grid on each poloidal slice using the elliptic grid generator

    nx = 20
    ny = 20

    pol_slices = [
        poloidal_grid.grid_elliptic(inner_line, outer_line, nx, ny, show=True)
        for inner_line in inner_lines
    ]



import boututils.calculus as calc
import numpy as np
from boututils.datafile import DataFile

import zoidberg as zb


# calculate curvature for curvilinear grids
def calc_curvilinear_curvature(fname, field, grid, maps):
    f = DataFile(str(fname), write=True)
    B = f.read("B")

    dx = grid.metric()["dx"]
    dz = grid.metric()["dz"]
    g_11 = grid.metric()["g_xx"]
    g_22 = grid.metric()["g_yy"]
    g_33 = grid.metric()["g_zz"]
    g_12 = 0.0
    g_13 = grid.metric()["g_xz"]
    g_23 = 0.0

    GR = np.zeros(B.shape)
    GZ = np.zeros(B.shape)
    Gphi = np.zeros(B.shape)
    dRdz = np.zeros(B.shape)
    dZdz = np.zeros(B.shape)
    dRdx = np.zeros(B.shape)
    dZdx = np.zeros(B.shape)

    for y in np.arange(0, B.shape[1]):
        pol, _ = grid.getPoloidalGrid(y)
        R = pol.R
        Z = pol.Z
        # G = \vec{B}/B, here in cylindrical coordinates
        GR[:, y, :] = field.Bxfunc(R, y, Z) / ((B[:, y, :]) ** 2)
        GZ[:, y, :] = field.Bzfunc(R, y, Z) / ((B[:, y, :]) ** 2)
        Gphi[:, y, :] = field.Byfunc(R, y, Z) / ((B[:, y, :]) ** 2)
        for x in np.arange(0, B.shape[0]):
            dRdz[x, y, :] = calc.deriv(R[x, :]) / dz[x, y, :]
            dZdz[x, y, :] = calc.deriv(Z[x, :]) / dz[x, y, :]
        for z in np.arange(0, B.shape[-1]):
            dRdx[:, y, z] = calc.deriv(R[:, z]) / dx[:, y, z]
            dZdx[:, y, z] = calc.deriv(Z[:, z]) / dx[:, y, z]

    R = f.read("R")
    Z = f.read("Z")
    dy = f.read("dy")

    # calculate Jacobian and contravariant terms in curvilinear coordinates
    J = R * (dZdz * dRdx - dZdx * dRdz)
    Gx = (GR * dZdz - GZ * dRdz) * (R / J)
    Gz = (GZ * dRdx - GR * dZdx) * (R / J)

    G_x = Gx * g_11 + Gphi * g_12 + Gz * g_13
    G_y = Gx * g_12 + Gphi * g_22 + Gz * g_23
    G_z = Gx * g_13 + Gphi * g_23 + Gz * g_33

    dG_zdy = np.zeros(B.shape)
    dG_ydz = np.zeros(B.shape)
    dG_xdz = np.zeros(B.shape)
    dG_zdx = np.zeros(B.shape)
    dG_ydx = np.zeros(B.shape)
    dG_xdy = np.zeros(B.shape)
    for y in np.arange(0, B.shape[1]):
        for x in np.arange(0, B.shape[0]):
            dG_ydz[x, y, :] = calc.deriv(G_y[x, y, :]) / dz[x, y, :]
            dG_xdz[x, y, :] = calc.deriv(G_x[x, y, :]) / dz[x, y, :]
        for z in np.arange(0, B.shape[-1]):
            dG_ydx[:, y, z] = calc.deriv(G_y[:, y, z]) / dx[:, y, z]
            dG_zdx[:, y, z] = calc.deriv(G_z[:, y, z]) / dx[:, y, z]

    # this should really use the maps...
    for x in np.arange(0, B.shape[0]):
        for z in np.arange(0, B.shape[-1]):
            dG_zdy[x, :, z] = calc.deriv(G_z[x, :, z]) / dy[x, :, z]
            dG_xdy[x, :, z] = calc.deriv(G_x[x, :, z]) / dy[x, :, z]

    bxcvx = (dG_zdy - dG_ydz) / J
    bxcvy = (dG_xdz - dG_zdx) / J
    bxcvz = (dG_ydx - dG_xdy) / J

    f.write("bxcvx", bxcvx)
    f.write("bxcvy", bxcvy)
    f.write("bxcvz", bxcvz)
    f.write("J", J)
    f.close()


def get_lines(
    field, start_r, start_z, yslices, yperiod=2 * np.pi, npoints=150, smoothing=False
):
    rzcoord, ycoords = zb.fieldtracer.trace_poincare(
        field, start_r, start_z, yperiod, y_slices=yslices, revs=npoints
    )

    lines = []
    for i in range(ycoords.shape[0]):
        r = rzcoord[:, i, 0, 0]
        z = rzcoord[:, i, 0, 1]
        line = zb.rzline.line_from_points(r, z)
        # Re-map the points so they're approximately uniform in distance along the surface
        # Note that this results in some motion of the line
        line = line.equallySpaced()
        lines.append(line)

    return lines


# smooth the metric tensor components
def smooth_metric(
    fname, write_to_file=False, return_values=False, smooth_metric=True, order=7
):
    from scipy.signal import savgol_filter

    f = DataFile(str(fname), write=True)
    bxcvx = f.read("bxcvx")
    bxcvz = f.read("bxcvz")
    bxcvy = f.read("bxcvy")
    J = f.read("J")

    bxcvx_smooth = np.zeros(bxcvx.shape)
    bxcvy_smooth = np.zeros(bxcvy.shape)
    bxcvz_smooth = np.zeros(bxcvz.shape)
    J_smooth = np.zeros(J.shape)

    if smooth_metric:
        g13 = f.read("g13")
        g_13 = f.read("g_13")
        g11 = f.read("g11")
        g_11 = f.read("g_11")
        g33 = f.read("g33")
        g_33 = f.read("g_33")

        g13_smooth = np.zeros(g13.shape)
        g_13_smooth = np.zeros(g_13.shape)
        g11_smooth = np.zeros(g11.shape)
        g_11_smooth = np.zeros(g_11.shape)
        g33_smooth = np.zeros(g33.shape)
        g_33_smooth = np.zeros(g_33.shape)

    for y in np.arange(0, bxcvx.shape[1]):
        for x in np.arange(0, bxcvx.shape[0]):
            bxcvx_smooth[x, y, :] = savgol_filter(
                bxcvx[x, y, :], np.int(np.ceil(bxcvx.shape[-1] / 2) // 2 * 2 + 1), order
            )
            bxcvz_smooth[x, y, :] = savgol_filter(
                bxcvz[x, y, :], np.int(np.ceil(bxcvz.shape[-1] / 2) // 2 * 2 + 1), order
            )
            bxcvy_smooth[x, y, :] = savgol_filter(
                bxcvy[x, y, :], np.int(np.ceil(bxcvy.shape[-1] / 2) // 2 * 2 + 1), order
            )
            J_smooth[x, y, :] = savgol_filter(
                J[x, y, :], np.int(np.ceil(J.shape[-1] / 2) // 2 * 2 + 1), order
            )
            if smooth_metric:
                g11_smooth[x, y, :] = savgol_filter(
                    g11[x, y, :], np.int(np.ceil(g11.shape[-1] / 2) // 2 * 2 + 1), order
                )
                g_11_smooth[x, y, :] = savgol_filter(
                    g_11[x, y, :],
                    np.int(np.ceil(g_11.shape[-1] / 2) // 2 * 2 + 1),
                    order,
                )
                g13_smooth[x, y, :] = savgol_filter(
                    g13[x, y, :], np.int(np.ceil(g13.shape[-1] / 2) // 2 * 2 + 1), order
                )
                g_13_smooth[x, y, :] = savgol_filter(
                    g_13[x, y, :],
                    np.int(np.ceil(g_13.shape[-1] / 2) // 2 * 2 + 1),
                    order,
                )
                g33_smooth[x, y, :] = savgol_filter(
                    g33[x, y, :], np.int(np.ceil(g33.shape[-1] / 2) // 2 * 2 + 1), order
                )
                g_33_smooth[x, y, :] = savgol_filter(
                    g_33[x, y, :],
                    np.int(np.ceil(g_33.shape[-1] / 2) // 2 * 2 + 1),
                    order,
                )

    if write_to_file:
        # f.write('bxcvx',bxcvx_smooth)
        # f.write('bxcvy',bxcvy_smooth)
        # f.write('bxcvz',bxcvz_smooth)
        f.write("J", J_smooth)

        if smooth_metric:
            f.write("g11", g11_smooth)
            f.write("g_11", g_11_smooth)
            f.write("g13", g13_smooth)
            f.write("g_13", g_13_smooth)
            f.write("g33", g33_smooth)
            f.write("g_33", g_33_smooth)

    f.close()
    if return_values:
        return bxcvx_smooth, bxcvy_smooth, bxcvz_smooth, bxcvx, bxcvy, bxcvz


def calc_iota(field, start_r, start_z):
    from scipy.signal import argrelextrema

    toroidal_angle = np.linspace(0.0, 400 * np.pi, 10000, endpoint=False)
    result = zb.fieldtracer.FieldTracer.follow_field_lines(
        field, start_r, start_z, toroidal_angle
    )
    peaks = argrelextrema(result[:, 0, 0], np.greater, order=10)[0]
    iota_bar = 2 * np.pi / (toroidal_angle[peaks[1]] - toroidal_angle[peaks[0]])
    # plt.plot(toroidal_angle, result[:,0,0]); plt.show()
    return iota_bar



import zoidberg as zb
import numpy as np
import xarray as xr
from mms_helper import lst

nonlin = 0.8
assert nonlin < 1


def var1(*args):
    pos = np.linspace(*args)
    return pos + np.sin(pos) * nonlin


def var2(*args):
    pos = np.linspace(*args)
    return pos + np.cos(pos) * nonlin


def var3(*args):
    pos = np.linspace(*args)
    return pos + np.cos(pos * 2) / 2 * nonlin


def var4(*args):
    pos = np.linspace(*args)
    return pos + np.sin(pos) / 2 * nonlin


modes = [
    ("const", np.linspace),
    ("var1", var1),
    ("var2", var2),
    ("var3", var3),
    ("var4", var4),
]


def gen_name(*args):
    nx, ny, nz, R0, r0, r1, mode = args
    return f"poloidal_{modes[mode][0]}_{nx}_{ny}_{nz}_{R0}.fci.nc"


def gen_grid(*args):
    nx, ny, nz, R0, r0, r1, mode = args
    mode = modes[mode]
    one = np.ones((nx, ny, nz))
    r = np.linspace(r0, r1, nx)[:, None]
    theta = mode[1](0, 2 * np.pi, nz, False)[None, :]
    phi = np.linspace(0, 2 * np.pi / 5, ny, False)
    R = R0 + np.cos(theta) * r
    Z = np.sin(theta) * r
    pol_grid = zb.poloidal_grid.StructuredPoloidalGrid(R, Z)

    field = zb.field.CurvedSlab(Bz=0, Bzprime=0, Rmaj=R0)
    grid = zb.grid.Grid(pol_grid, phi, 5, yperiodic=True)

    fn = gen_name(*args)

    maps = zb.make_maps(grid, field, quiet=True, MXG=1)
    zb.write_maps(
        grid,
        field,
        maps,
        "tmp.nc",
        metric2d=False,
    )
    with xr.open_dataset("tmp.nc") as ds:
        dims = ds.dz.dims
        ds["r_minor"] = dims, one * r[:, None, :]
        ds["phi"] = "y", phi
        ds["theta"] = dims, one * theta[:, None, :]
        ds["one"] = dims, one
        ds["g12"] = dims, one * 0
        ds["g23"] = dims, one * 0
        ds["g_12"] = dims, one * 0
        ds["g_23"] = dims, one * 0
        ds.to_netcdf(fn)


def _togen(*args):
    return gen_name(*args), args


grids = {}
for mode in range(5):
    grids[modes[mode][0]] = [_togen(4, 2, nz, 1, 0.1, 0.5, mode) + (nz,) for nz in lst]

if __name__ == "__main__":
    for todos in grids.values():
        for fn, args, _ in todos:
            gen_grid(*args)



import boutpp as bc
import xarray as xr
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np

lst = [4, 8, 16, 32, 64][:-1]


def extend(p):
    return np.concatenate([p[:, -1:, :], p, p[:, :1, :]], axis=1)


def diff(p, axis):
    if axis == 0:
        out = np.zeros_like(p)
        out[:-1] = p[1:] - p[:-1]
        out[1:] += p[1:] - p[:-1]
        out[1:-1] /= 2
        return out
    return np.roll(p, 1, axis=axis) - p


def diff2(p, axis):
    assert axis != 0
    return (2 * p - np.roll(p, 1, axis=axis) - np.roll(p, -1, axis=axis)) / 2


def clean(p):
    return p[:, 1:-1, :]


def fixup(fn):
    grid = xr.open_dataset(fn)
    R = grid.R
    Z = grid.Z
    dims = Z.dims
    one = np.ones_like(Z)

    # grid["dx"] = dims, np.sqrt((diff(R, 0)) ** 2 + (diff(Z, 0)) ** 2)
    # grid.dx[:, 1, :].plot()
    # plt.figure()
    dxn = np.sqrt((diff(R, 0)) ** 2 + (diff(Z, 0)) ** 2)
    dxn /= np.mean(dxn) * R.shape[0]

    # plt.imshow(dxn[:, 1])
    # plt.colorbar()
    # plt.show()

    dzn = np.sqrt((diff(R, 2)) ** 2 + (diff(Z, 2)) ** 2)
    dzn /= np.mean(dzn) * R.shape[2] / 2 / np.pi
    # grid["dz"] = np.sqrt((diff(R, 2)) ** 2 + (diff(Z, 2)) ** 2)
    ny = R.shape[1]
    dy = 2 * np.pi / 5 / ny
    phi = np.linspace(0, np.pi / 5, ny * 5, False)[None, :ny, None] * one + dy / 2
    # grid["dy"] = dims, dy * one

    # grid["dx"] = dims, dxn
    # grid["dy"] = dims, dy * one
    # grid["dz"] = dzn
    # grid["phi"] = (dims, phi)

    # J = np.array([[diff(o, d) for d in range(3)] for o in [R, Z, phi]])

    # g = np.array(
    #     [[[J[k, i] * J[k, j] for k in range(3)] for i in range(3)] for j in range(3)]
    # )
    # g = g.sum(axis=2)
    # g = g.transpose((2, 3, 4, 0, 1))
    # J = J.transpose((2, 3, 4, 0, 1))
    # # g
    # assert np.all(J[..., 0, 1] == diff(R, 1))
    # # Jinv = np.linalg.inv(J)
    # bad = np.linalg.det(g) == 0
    # for i, b in enumerate(bad):
    #     if np.any(b):
    #         plt.figure()
    #         # plt.imshow(np.linalg.det(g)[i])
    #         plt.imshow(b)
    #         plt.colorbar()
    #         plt.title(str(i))
    # plt.show()

    # #    raise
    # ginv = np.linalg.inv(g)
    # det = np.linalg.det(J)
    # if np.min(det) * np.max(det) < 0:
    #     plt.imshow(g[:, 2, :, 0, 0])
    #     plt.colorbar()
    #     plt.show()

    # for i in range(3):
    #     for j in range(3):
    #         grid[f"g_{i+1}{j+1}"] = dims, g[..., i, j]
    #         grid[f"g{i+1}{j+1}"] = dims, ginv[..., i, j]
    g = np.empty((3, 3), dtype=object)
    for i in range(3):
        for j in range(3):
            k = f"g{i+1}{j+1}"
            if k in grid:
                g[i, j] = grid[k]
            else:
                g[i, j] = 0 * one
    g = np.array([list(y) for y in g]).transpose((2, 3, 4, 0, 1))
    assert np.all(np.linalg.det(g) > 0)
    for _ in range(6):
        d = diff2(g, 2) + diff2(g, 1)
        fac = 10
        if not np.all(np.linalg.det(g - d / fac) > 0):
            fac = fac * 2
        print(fac)
        g -= d / fac
    assert np.all(np.linalg.det(g) > 0)
    ginv = np.linalg.inv(g)
    # print(g.shape)

    # raise
    # for k in f"g_{i+1}{j+1}", :
    #     if k not in grid:
    #         grid[k] = dims, 0 * one

    for i in range(3):
        for j in range(3):
            for k, d in zip([f"g_{i+1}{j+1}", f"g{i+1}{j+1}"], [ginv, g]):
                grid[k] = grid.g11.dims, d[..., i, j]
                # grid[k] -= diff2(grid[k].data, 2) / 10
    # Force recalculation
    del grid["J"]
    # grid["J"] = grid.g11.dims, np.linalg.det(g)
    # grid.J.data[grid.J.data < 0] = 0
    # for i in range(3):
    #     grid.J.data -= diff2(grid.J.data, 2) / 10
    #     # plt.plot(grid[k][2, 2], "r")
    # # plt.show()
    # assert np.all(grid.J > 0)

    grid.to_netcdf(f"{fn[:-3]}.fix.nc")


def test(fn):
    grid = xr.open_dataset(fn)
    bc.Options().set("mesh:file", fn, force=True)
    mesh = bc.Mesh(section="")
    print(bc.__file__)
    print()
    print()
    print()
    print()
    f = bc.create3D("0", mesh)
    R = extend(grid.R)
    Z = extend(grid.Z)
    one = np.ones_like(Z)

    dx = np.sqrt((diff(R, 0)) ** 2 + (diff(Z, 0)) ** 2)
    dz = np.sqrt((diff(R, 2)) ** 2 + (diff(Z, 2)) ** 2)
    ny = R.shape[1] - 2
    dy = 2 * np.pi / 5 / ny
    phi = np.linspace(0, np.pi / 5, ny * 5, False)[None, : ny + 2, None] - dy * 1.5

    coords = mesh.coordinates
    print(dx.shape)
    # coords.dx[:, :, :] = dx
    # coords.dy[:, :, :] = dy * one
    # coords.dz[:, :, :] = dz

    X = R * np.cos(phi)
    Y = R * np.sin(phi)
    inp = np.sin(X)  # * np.sin(Z) + np.cos(Y)
    # ana = -2 * np.sin(X) * np.sin(Z) - np.cos(Y)
    ana = -np.sin(X)

    # Set input field
    f[:, :, :] = inp
    mesh.communicate(f)
    calc = bc.Laplace(f).get()
    l2 = np.sqrt(np.mean(clean(ana - calc)[2:-2, 1:-1] ** 2))
    ## print(fn, l2)
    #
    fig, axs = plt.subplots(1, 3)
    calcd = clean(calc)[:, 1]
    anad = ana[:, 1]
    for dat, label, ax in zip(
        [anad, calcd, (calcd - anad)], ["ana", "calc", "err"], axs
    ):
        plot = ax.imshow(dat[2:-2, 1:-1])
        ax.set_title(fn[-20:] + " " + label)
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        plt.colorbar(plot, cax=cax, orientation="vertical")
    # plt.figure()
    # plt.imshow((calcd - anad)[1:-1])
    # plt.title(fn + " err")
    # plt.colorbar()
    return l2


if __name__ == "__main__":
    for x in lst[0:]:
        fixup(f"rotating-ellipse.{x}x{x}x{x}.fci.nc")

    bc.init("-d mms -q -q -q")

    l2 = [test(f"rotating-ellipse.{x}x{x}x{x}.fci.fix.nc") for x in lst[-3:]]

    print(l2)

    plt.show()
# for x in lst:



from mms_helper import extend, do_tests, plt, bc, np
from radial_grid import grids


set1 = (
    lambda grid: np.sin(extend(grid.r_minor)),
    lambda grid: np.cos(extend(grid.r_minor)),
    "sin(r)",
)

set2 = (
    lambda grid: extend(grid.r_minor) ** 3,
    lambda grid: 3 * extend(grid.r_minor) ** 2,
    "r³",
)


if __name__ == "__main__":
    bc.init("-d mms -q -q -q")

    do_tests(grids, *set1, bc.DDX)
    do_tests(grids, *set2, bc.DDX)

    plt.show()



#!/usr/bin/env python

# Create grids for a straight stellarator, based on
# curvilinear grids

import numpy as np
import matplotlib.pyplot as plt

import zoidberg

#############################################################################
# Define the magnetic field

# Length in y after which the coils return to their starting (R,Z) locations
yperiod = 10.0

magnetic_field = zoidberg.field.StraightStellarator(
    I_coil=0.4, radius=1.0, yperiod=yperiod
)

zoidberg.plot.plot_poincare(magnetic_field, 0.3, 0.0, yperiod)



#!/usr/bin/env python

import numpy as np
import zoidberg

field = zoidberg.field.GEQDSK("test_1.geqdsk")  # Read magnetic field

grid = zoidberg.grid.rectangular_grid(
    100,
    10,
    100,
    1.5 - 0.1,  # Range in R (max - min)
    2 * np.pi,  # Toroidal angle
    3.0,  # Range in Z
    xcentre=(1.5 + 0.1) / 2,  # Middle of grid in R
    yperiodic=True,
)  # Periodic in toroidal angle

# Create the forward and backward maps
maps = zoidberg.make_maps(grid, field)

# Save to file
zoidberg.write_maps(grid, field, maps, gridfile="tokamak.fci.nc")

#############################################################################
# Plot maps

zoidberg.plot.plot_forward_map(grid, maps)



from mms_helper import extend, do_tests, plt, bc, np
from delp_grid import grids


sets = [
    # (
    #     lambda grid: np.sin(extend(grid["one"] * grid.phi) * 5),
    #     lambda grid: -25 * np.sin(extend(grid["one"] * grid.phi) * 5),
    # ),
    (
        lambda grid: np.sin(extend(grid.Z)),
        lambda grid: -np.sin(extend(grid.Z)),
        "sin(Z)",
    ),
    (
        lambda grid: np.sin(extend(grid.R)),
        lambda grid: -np.sin(extend(grid.R)) + np.cos(extend(grid.R)) / extend(grid.R),
        "sin(R)",
    ),
]
# set2 = (
#     lambda grid: np.sin(extend(grid["one"] * grid.phi + grid.r_minor + grid.theta) * 5)
#     * extend(np.cos(grid.theta)),
#     lambda grid: 5
#     * np.cos(extend(grid["one"] * grid.phi + grid.r_minor + grid.theta) * 5)
#     * extend(np.cos(grid.theta)),
# )


if __name__ == "__main__":
    bc.init("-d mms -q -q -q")

    for seti in sets:
        do_tests(grids, *seti, bc.Laplace)
    # do_tests(grids, *set2, bc.DDX)

    plt.show()



"""Boundary objects that define an 'outside'

"""

import numpy as np


class NoBoundary(object):
    """No boundary, so no points outside"""

    def outside(self, x, y, z):
        """Returns True if the point is outside the boundary

        Parameters
        ----------
        x, y, z : array_like
            Coordinates of the point(s) to check

        Returns
        -------
        bool
            True if point is outside boundary
        """
        return np.full(x.shape, False, dtype=bool)


class RectangularBoundaryXZ(object):
    def __init__(self, xmin, xmax, zmin, zmax):
        """Create a simple rectangular boundary box

        Parameters
        ----------
        xmin, xmax : float
            The X coordinates of the limits of the box
        zmin, zmax : float
            The Z coordinates of the limits of the box
        """

        self.xmin = xmin
        self.xmax = xmax
        self.zmin = zmin
        self.zmax = zmax

    def outside(self, x, y, z):
        """Returns true if the given point is outside the domain

        Parameters
        ----------
        x, y, z : array_like
            Coordinates of the point(s) to check

        Returns
        -------
        bool
            True if point is outside boundary

        """

        return np.logical_or(
            np.logical_or(x < self.xmin, x > self.xmax),
            np.logical_or(z < self.zmin, z > self.zmax),
        )


class PolygonBoundaryXZ(object):
    def __init__(self, xarr, zarr):
        """Create a polygon boundary defined by

        (xarr[0], zarr[0]) -- (xarr[1], zarr[1]) -- ...

        Parameters
        ----------
        xarr, zarr : array_like
            The X, Z coordinates of the polygon vertices
        """

        # Find a point outside the polygon
        self.zout = np.amax(zarr) + 1.0  # Somewhere outside the polygon
        self.xarr = xarr
        self.zarr = zarr

    def outside(self, x, y, z):
        """Returns true if the given point is outside the domain

        Parameters
        ----------
        x, y, z : array_like
            Coordinates of the point(s) to check

        Returns
        -------
        bool
            True if point is outside boundary

        """
        # Find intersection of a line from z = z_0 to z_1
        #
        #    x = x_0    z = (1-a)*z_0 + a*z_1
        #
        # with a line segment (x_0, z_0) to (x_1, z_1)
        #
        #    x' = (1-b)*x'_0 + b*x'_1   z' = (1-b)*z'_0 + b*z'_1
        #
        # where a and b must be between 0 and 1 for an intersection
        #

        x_0 = x
        z_0 = z
        z_1 = self.zout

        # Create an array of True values, marking all points
        # as outside the domain
        outside = np.full(x.shape, True, dtype=bool)

        # Each intersection with a boundary segment
        # changes True->False->True i.e. XOR

        nsegments = len(self.xarr)
        for i in range(nsegments):  # Loop over all segments
            xp_0 = self.xarr[i]
            zp_0 = self.zarr[i]

            ip = (i + 1) % nsegments  # Next point
            xp_1 = self.xarr[ip]
            zp_1 = self.zarr[ip]

            # First calculate b = (x_0 - x'_0) / (x'_1 - x'_0)
            # but check for case where (x'_1 - x'_0) is small
            if abs(xp_1 - xp_0) < 1e-6:
                continue  # Skip this segment

            b = (x_0 - xp_0) / (xp_1 - xp_0)

            a = (zp_0 - z_0 + b * (zp_1 - zp_0)) / (z_1 - z_0)

            intersect = np.logical_and(
                np.logical_and(b >= 0.0, b <= 1.0), np.logical_and(a >= 0.0, a <= 1.0)
            )

            # intersect is now true for those points which intersect this line segment
            outside = np.logical_xor(outside, intersect)

        return outside



#!/usr/bin/env python

# Demonstrates how to generate Poincare plots of magnetic fields

import zoidberg
import numpy as np

# Size of the domain in y (periodic)
yperiod = 10.0

# Define magnetic field
magnetic_field = zoidberg.field.StraightStellarator(
    I_coil=0.4, radius=1.0, yperiod=yperiod
)

# Make Poincare plot
zoidberg.plot.plot_poincare(
    magnetic_field,
    np.linspace(0, 0.5, 5),  # x starting positions
    0.0,  # z starting positions
    yperiod,  # Periodicity of y domain
    interactive=True,
)  # Click on plot to add points



import numpy as np

from .grid import rectangular_grid


def test_getPoloidalGrid():
    # Create a simple non-periodic grid
    grid = rectangular_grid(10, 10, 10, yperiodic=False)

    # Check number of y points
    assert grid.numberOfPoloidalGrids() == 10

    # Try to get points outside the domain
    p, y = grid.getPoloidalGrid(-1)
    assert p is None
    assert np.allclose(y, 0.0)

    p, y = grid.getPoloidalGrid(10)
    assert p is None
    assert np.allclose(y, 10.0)

    # Check values of y at the start and end of the domain
    p, y = grid.getPoloidalGrid(0)
    assert np.allclose(y, 0.5)

    p, y = grid.getPoloidalGrid(9)
    assert np.allclose(y, 9.5)


def test_getPoloidalGrid_periodic():
    # Create a periodic grid
    grid = rectangular_grid(10, 10, 10, yperiodic=True)

    assert grid.numberOfPoloidalGrids() == 10

    p_last, y_last = grid.getPoloidalGrid(9)  # Last in domain

    assert p_last is not None

    p, y = grid.getPoloidalGrid(-1)

    assert p is p_last
    assert y < y_last



import argparse

import numpy as np
from boututils.datafile import DataFile

import zoidberg as zb

from .common import (calc_curvilinear_curvature, calc_iota, get_lines,
                     smooth_metric)


def rotating_ellipse(
    nx=68,
    ny=16,
    nz=128,
    xcentre=5.5,
    I_coil=0.01,
    curvilinear=True,
    rectangular=False,
    fname="rotating-ellipse.fci.nc",
    a=0.4,
    curvilinear_inner_aligned=True,
    curvilinear_outer_aligned=True,
    npoints=2467,
    Btor=2.5,
    show_maps=False,
    calc_curvature=True,
    smooth_curvature=False,
    return_iota=True,
    write_iota=False,
):
    yperiod = 2 * np.pi / 5.0
    field = zb.field.RotatingEllipse(
        xcentre=xcentre, I_coil=I_coil, radius=2 * a, yperiod=yperiod, Btor=Btor
    )
    # Define the y locations
    ycoords = np.linspace(0.0, yperiod, ny, endpoint=False)
    start_r = xcentre + a / 2.0
    start_z = 0.0

    if rectangular:
        print("Making rectangular poloidal grid")
        poloidal_grid = zb.poloidal_grid.RectangularPoloidalGrid(
            nx, nz, 1.0, 1.0, Rcentre=xcentre
        )
    elif curvilinear:
        print("Making curvilinear poloidal grid")
        inner = zb.rzline.shaped_line(
            R0=xcentre, a=a / 2.0, elong=0, triang=0.0, indent=0, n=npoints
        )
        outer = zb.rzline.shaped_line(
            R0=xcentre, a=a, elong=0, triang=0.0, indent=0, n=npoints
        )

        if curvilinear_inner_aligned:
            print("Aligning to inner flux surface...")
            inner_lines = get_lines(
                field, start_r, start_z, ycoords, yperiod=yperiod, npoints=npoints
            )
        if curvilinear_outer_aligned:
            print("Aligning to outer flux surface...")
            outer_lines = get_lines(
                field, xcentre + a, start_z, ycoords, yperiod=yperiod, npoints=npoints
            )

        print("creating grid...")
        if curvilinear_inner_aligned:
            if curvilinear_outer_aligned:
                poloidal_grid = [
                    zb.poloidal_grid.grid_elliptic(inner, outer, nx, nz, show=show_maps)
                    for inner, outer in zip(inner_lines, outer_lines)
                ]
            else:
                poloidal_grid = [
                    zb.poloidal_grid.grid_elliptic(inner, outer, nx, nz, show=show_maps)
                    for inner in inner_lines
                ]
        else:
            poloidal_grid = zb.poloidal_grid.grid_elliptic(inner, outer, nx, nz)

    # Create the 3D grid by putting together 2D poloidal grids
    grid = zb.grid.Grid(poloidal_grid, ycoords, yperiod, yperiodic=True)
    maps = zb.make_maps(grid, field)
    zb.write_maps(grid, field, maps, str(fname), metric2d=False)

    if curvilinear and calc_curvature:
        print("calculating curvature...")
        calc_curvilinear_curvature(fname, field, grid, maps)

    if calc_curvature and smooth_curvature:
        smooth_metric(
            fname, write_to_file=True, return_values=False, smooth_metric=True
        )

    if return_iota or write_iota:
        iota_bar = calc_iota(field, start_r, start_z)
        if write_iota:
            f = DataFile(str(fname), write=True)
            f.write("iota_bar", iota_bar)
            f.close()
        else:
            print("Iota_bar = ", iota_bar)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--nx", "-x", help="number of x points", default=68, type=int)
    parser.add_argument("--ny", "-y", help="number of y points", default=16, type=int)
    parser.add_argument("--nz", "-z", help="number of z points", default=128, type=int)
    parser.add_argument(
        "--xcentre", "-c", help="Centre of the screw pinch", default=5.5, type=float
    )
    parser.add_argument("--I-coil", "-i", help="Coil current", default=0.01, type=float)
    parser.add_argument("--curvilinear", help="???", action="store_false")
    parser.add_argument("--rectangular", help="???", action="store_true")
    parser.add_argument("--curvilinear_outer_aligned", help="???", action="store_false")
    parser.add_argument("--curvilinear_inner_aligned", help="???", action="store_false")
    parser.add_argument(
        "--fname",
        "-f",
        help="File name of the grid to be written",
        default="rotating-ellipse.fci.nc",
        type=str,
    )
    parser.add_argument(
        "--Btor", "-b", help="magnetic field on axis", default=2.5, type=float
    )
    parser.add_argument("-a", help="???", default=0.4, type=float)
    parser.add_argument(
        "--npoints",
        "-p",
        help="Number of iterations for surface tracing. Needs to be large enough.",
        default=2467,
        type=int,
    )

    parser.add_argument("--calc_curvature", help="???", action="store_false")
    parser.add_argument("--smooth_curvature", help="???", action="store_true")
    parser.add_argument(
        "--write_iota", help="Include iota in the grid file", action="store_true"
    )
    parser.add_argument("--show_maps", "-s", help="plot the grid", action="store_true")

    args = parser.parse_args()
    rotating_ellipse(return_iota=False, **args.__dict__)


if __name__ == "__main__":
    main()



import numpy as np
import pytest

from . import field, fieldtracer


def test_slab():
    mag_field = field.Slab(By=0.5, Bz=0.1, Bzprime=0.0)

    tracer = fieldtracer.FieldTracer(mag_field)

    coords = tracer.follow_field_lines(0.2, 0.3, [1.5, 2.5])

    assert coords.shape == (2, 1, 2)
    assert np.allclose(coords[:, 0, 0], 0.2)  # X coordinate

    assert np.allclose(coords[0, 0, 1], 0.3)
    assert np.allclose(coords[1, 0, 1], 0.3 + 0.1 / 0.5)  # Z coordinate

    coords = tracer.follow_field_lines([0.2, 0.3, 0.4], 0.5, [1.0, 2.5])

    assert coords.shape == (2, 3, 2)
    assert np.allclose(coords[:, 0, 0], 0.2)
    assert np.allclose(coords[:, 1, 0], 0.3)
    assert np.allclose(coords[:, 2, 0], 0.4)
    assert np.allclose(coords[1, :, 1], 0.5 + 1.5 * 0.1 / 0.5)  # Z coord


def test_FieldTracerReversible_slab():
    mag_field = field.Slab(By=0.5, Bz=0.1, Bzprime=0.0)

    tracer = fieldtracer.FieldTracerReversible(mag_field)

    coords = tracer.follow_field_lines(0.2, 0.3, [1.5, 2.5])

    assert coords.shape == (2, 1, 2)
    assert np.allclose(coords[:, 0, 0], 0.2)  # X coordinate

    assert np.allclose(coords[0, 0, 1], 0.3)
    assert np.allclose(coords[1, 0, 1], 0.3 + 0.1 / 0.5)  # Z coordinate

    coords = tracer.follow_field_lines([0.2, 0.3, 0.4], 0.5, [1.0, 2.5])

    assert coords.shape == (2, 3, 2)
    assert np.allclose(coords[:, 0, 0], 0.2)
    assert np.allclose(coords[:, 1, 0], 0.3)
    assert np.allclose(coords[:, 2, 0], 0.4)
    assert np.allclose(coords[1, :, 1], 0.5 + 1.5 * 0.1 / 0.5)  # Z coord


def test_poincare():
    mag_field = field.Slab(By=0.5, Bz=0.1, Bzprime=0.0)

    result, y_slices = fieldtracer.trace_poincare(
        mag_field, 0.0, 0.0, 1.0, nplot=3, revs=5, nover=1
    )

    assert y_slices.size == 3
    assert result.shape == (5, y_slices.size, 1, 2)


def setup_ftw(**kw):
    try:
        import osa
    except ImportError:
        pytest.skip("osa not installed")
    import requests

    try:
        return fieldtracer.FieldTracerWeb(**kw)
    except requests.exceptions.ReadTimeout:
        pytest.skip("Failed to initiallise - service not available?")


def test_traceweb():
    web = setup_ftw(configId=0)

    num = 10
    start = np.linspace(5.7, 6, num), np.zeros(num)

    res = web.follow_field_lines(*start, np.linspace(0, 0.1, 10))
    assert np.all(res[1:, :, 1] > 0)
    dist = np.sqrt(np.sum((res[:, 1:, :] - res[:, :-1, :]) ** 2, axis=2))
    assert np.all(dist >= 0)
    assert np.max(dist) < 0.05

    # def test_traceweb_backwards():
    #     web = setup_ftw(configId=0)

    #     num = 10
    #     start = np.linspace(5.7, 6, num), np.zeros(num)

    # Check reverse direction
    res = web.follow_field_lines(*start, -np.linspace(0, 0.1, 10))
    assert np.all(res[1:, :, 1] < 0)
    dist = np.sqrt(np.sum((res[:, 1:, :] - res[:, :-1, :]) ** 2, axis=2))
    assert np.all(dist >= 0)
    # make sure step isn't to large
    assert np.max(dist) < 0.05

    # def test_traceweb_forward_phi0():
    #     web = setup_ftw(configId=0)

    #     num = 10
    #     start = np.linspace(5.7, 6, num), np.zeros(num)

    res = web.follow_field_lines(*start, np.linspace(0, 0.1, 10) + np.pi * 4 / 5)
    assert np.all(res[1:, :, 1] > 0)
    dist = np.sqrt(np.sum((res[:, 1:, :] - res[:, :-1, :]) ** 2, axis=2))
    assert np.all(dist >= 0)
    assert np.max(dist) < 0.05

    # def test_traceweb_chunked():
    #     web = setup_ftw(configId=0)

    #     num = 10
    #     start = np.linspace(5.7, 6, num), np.zeros(num)

    res = web.follow_field_lines(*start, -np.linspace(0, 0.1, 10) + np.pi * 6 / 5)
    assert np.all(res[1:, :, 1] < 0)
    dist = np.sqrt(np.sum((res[:, 1:, :] - res[:, :-1, :]) ** 2, axis=2))
    assert np.all(dist >= 0)
    assert np.max(dist) < 0.05

    web = fieldtracer.FieldTracerWeb(configId=0, chunk=2)
    res_chunked = web.follow_field_lines(
        *start, -np.linspace(0, 0.1, 10) + np.pi * 6 / 5
    )
    assert np.allclose(res, res_chunked)


if __name__ == "__main__":
    test_traceweb()



import numpy as np
from scipy import linalg
from scipy.optimize import least_squares


def find_outliers(r, z, cutoff=0.75, bounds=10):
    ### find points in distance of 10 from each other, which create a too sharp pseudoangle
    outliers = []
    ### go through points with 10 distance
    for item in range(len(z)):
        point1 = np.array([z[item], r[item]])
        point2 = np.array([z[item - bounds], r[item - bounds]])
        point3 = np.array([z[item - bounds * 2], r[item - bounds * 2]])
        ### calculate pseudoangle without artan
        a = point1 - point2
        b = point1 - point3
        product = np.dot(a, b)
        magn_a = np.sqrt(a.dot(a))
        magn_b = np.sqrt(b.dot(b))
        pseudoangle = product / (magn_a * magn_b)
        if pseudoangle < cutoff:
            outliers.append(item - bounds)
    return outliers


def generate_points_tosmooth(bounds, outliers, length):
    ### for every position to be smoothed, get all points within a range surrouding that point
    # assert len(outliers) != 0
    if not outliers:
        return []
    todel = set()
    for dup in outliers:
        todel.update([x for x in range(dup - bounds, dup + bounds + 1) if x < length])
    ###
    todel = sorted(list(todel))
    splitedlist = []
    first = 0
    last = todel[0]
    for index, value in enumerate(todel[1:]):
        if last + 1 != value:
            splitedlist.append(todel[first : index + 1])
            first = index + 1
        last = value
    splitedlist.append(todel[first:])
    return splitedlist


def rollup(original_array, startindex, length, replacement):
    original_array = np.roll(original_array, -startindex)
    original_array[0:length] = replacement
    original_array = np.roll(original_array, startindex)
    return original_array


def gen_newr(r, z, splitedlist, bounds, plot):
    newr = r.copy()
    newz = z.copy()
    have_failed = False
    deb = []
    for alist in splitedlist:
        assert len(alist) >= bounds
        if len(alist) >= len(r) / 2:
            raise ValueError("Too many points to smooth")
        start_index = alist[0]
        end_index = alist[-1]
        diff_minus_r = r[start_index] - r[start_index - 1]
        diff_minus_z = z[start_index] - z[start_index - 1]
        if end_index < len(r) - 1:
            seconds = end_index + 1
        else:
            seconds = 0
        diff_plus_r = r[end_index] - r[seconds]
        diff_plus_z = z[end_index] - z[seconds]
        fac = 1
        if 0:
            A = np.array(
                [
                    [1, 0.5, 1 / 3 / fac, 1 / 4, 0, 0, 0, 0],
                    [0, 0, 0, 0, 1, 0.5, 1 / 3 / fac, 1 / 4],
                    [diff_minus_z, 0, 0, 0, -diff_minus_r, 0, 0, 0],
                    [
                        diff_plus_z,
                        diff_plus_z,
                        diff_plus_z / fac,
                        diff_plus_z,
                        -diff_plus_r,
                        -diff_plus_r,
                        -diff_plus_r / fac,
                        -diff_plus_r,
                    ],
                ]
            )
        else:
            A = np.array(
                [
                    [1, 0.5, 1 / 3 / fac, 0, 0, 0],
                    [0, 0, 0, 1, 0.5, 1 / 3 / fac],
                    [diff_minus_z, 0, 0, -diff_minus_r, 0, 0],
                    [
                        diff_plus_z,
                        diff_plus_z,
                        diff_plus_z / fac,
                        -diff_plus_r,
                        -diff_plus_r,
                        -diff_plus_r / fac,
                    ],
                ]
            )

        b = np.array(
            [r[end_index] - r[start_index], z[end_index] - z[start_index], 0, 0]
        )

        def lenofs(x):
            if len(x) == 6:
                x1r, x2r, x3r = x[:3]
                x1z, x2z, x3z = x[3:]
                xr = x[:3]
                xz = x[3:]
                xs = xr * xr + xz * xz
                x1s, x2s, x3s = xs
                x12 = x1r * x2r + x1z * x2z
                x13 = x1r * x3r + x1z * x3z
                x23 = x2r * x3r + x2z * x3z
                num = 100
                s = np.linspace(0, 1, num)
                tmp = np.sqrt(
                    x1s
                    + x2s * s**2
                    + x3s * s**4
                    + 2 * (x12 * s + x13 * s**2 + x23 * s**3)
                )
                return np.sum(tmp) / num
            x1r, x2r, x3r, x4r = x[:4]
            x1z, x2z, x3z, x4z = x[4:]
            xr = x[:4]
            xz = x[4:]
            xs = xr * xr + xz * xz
            x1s, x2s, x3s, x4s = xs
            x12 = x1r * x2r + x1z * x2z
            x13 = x1r * x3r + x1z * x3z
            x14 = x1r * x4r + x1z * x4z
            x23 = x2r * x3r + x2z * x3z
            x24 = x2r * x4r + x2z * x4z
            x34 = x3r * x4r + x3z * x4z
            # return x1s + x12 + 2 / 3 * x13 + 1/2*x14 +1/3 * x23 + 1 / 4 * x2s + 1 / 5 * x3s +1/7*x4s
            num = 100
            s = np.linspace(0, 1, num)
            tmp = np.sqrt(
                x1s
                + x2s * s**2
                + x3s * s**4
                + x4s * s**6
                + 2
                * (
                    x12 * s
                    + x13 * s**2
                    + x14 * s**3
                    + x23 * s**3
                    + x24 * s**4
                    + x34 * s**5
                )
            )
            return np.sum(tmp) / num

        def mysin(a, b):
            aa, ab = a
            ba, bb = b
            return (aa * bb - ab * ba) / np.sqrt(
                (aa * aa + ab * ab + 1e-8) * (ba * ba + bb * bb + 1e-8)
            )

        def mycos(a, b):
            aa, ab = a
            ba, bb = b
            return (aa * ba + ab * bb) / np.sqrt(
                (aa * aa + ab * ab + 1e-8) * (ba * ba + bb * bb + 1e-8)
            )

        def curv(x, f):
            k = len(x) // 2
            a = np.sum(x[:k] * np.arange(k))
            b = x[1]
            c = np.sum(x[k:] * np.arange(k))
            d = x[k + 1]
            return (a * a + c * c) * f, (b * b + d * d) * f

        def fun(x, A, b, signs):
            ret = A @ x - b
            k = len(x) // 2
            newsigns = (x[0], x[1], np.sum(x[:k]), np.sum(x[k:]))
            mycoss = mycos(signs[:2], newsigns[:2]), mycos(signs[2:], newsigns[2:])
            return (
                *ret,
                lenofs(x) / 1e5,
                # (1 - mycoss[0]) / 1e2,
                # (mycoss[1] - 1) / 1e2,
                *(curv(x, 1 / 1e3)),
            )

        if 0:
            x = linalg.lstsq(A, b)
            x0 = x[0]
        else:
            signs = (diff_minus_r, diff_minus_z, -diff_plus_r, -diff_plus_z)
            x = least_squares(
                fun,
                np.zeros(A.shape[1]),
                args=(A, b, signs),
                gtol=None,
                xtol=1e-14,
                ftol=None,
            )
            # if np.max(np.abs(x.fun)) > 1e-5:
            if x.optimality > 1e-8:
                have_failed = True
                deb += [[x.cost, x.optimality, np.max(np.abs(x.fun))]]
                continue
            x0 = x.x
            newsigns = (x0[0], x0[1], np.sum(x0[:3]), np.sum(x0[3:]))

        S = np.linspace(0, 1, num=len(alist))

        if len(x0) == 6:
            X = (
                r[start_index]
                + x0[0] * S
                + (x0[1] * S**2) * 0.5
                + (x0[2] * S**3) / 3 / fac
            )
            Y = (
                z[start_index]
                + x0[3] * S
                + (x0[4] * S**2) * 0.5
                + (x0[5] * S**3) / 3 / fac
            )
        else:
            X = (
                r[start_index]
                + x0[0] * S
                + (x0[1] * S**2) * 0.5
                + (x0[2] * S**3) / 3 / fac
                + x0[3] * S**4 / 4
            )
            Y = (
                z[start_index]
                + x0[4] * S
                + (x0[5] * S**2) * 0.5
                + (x0[6] * S**3) / 3 / fac
                + x0[7] * S**4 / 4
            )

        # Check spacing at beginning and end
        dstart = (X[0] - newr[start_index - 1]) ** 2 + (
            Y[0] - newz[start_index - 1]
        ) ** 2
        dend = (X[-1] - newr[start_index + len(alist)]) ** 2 + (
            Y[-1] - newz[start_index + len(alist)]
        ) ** 2
        maxfac = 3
        if not (maxfac**-2 < dstart / dend < maxfac**2):
            print(f"dstart / dend is {dstart/dend} - check has failed")
            deb += [[dstart, dend, 3]]
            have_failed = True

        if start_index < 0:
            newr = rollup(newr, start_index, len(alist), X)
            newz = rollup(newz, start_index, len(alist), Y)
        else:
            newr[start_index : start_index + len(alist)] = X
            newz[start_index : start_index + len(alist)] = Y
        if plot:
            from matplotlib.collections import LineCollection

            S = np.linspace(0, 1, 1000)

            X = (
                r[start_index]
                + x0[0] * S
                + (x0[1] * S**2) * 0.5
                + (x0[2] * S**3) / 3 / fac
            )
            Y = (
                z[start_index]
                + x0[3] * S
                + (x0[4] * S**2) * 0.5
                + (x0[5] * S**3) / 3 / fac
            )
            points = np.array([X, Y]).T.reshape(-1, 1, 2)
            segments = np.concatenate([points[:-1], points[1:]], axis=1)
            norm = plt.Normalize(S.min(), S.max())
            lc = LineCollection(segments, cmap="viridis", norm=norm)
            lc.set_array(S)
            lc.set_linewidth(2)
            line = plt.gca().add_collection(lc)
            # plt.plot(X, Y, "k-")

    return newr, newz, have_failed, deb


alldeb = []


def smooth(r, z, cutoff=0.75, bounds=10, bounds_find=None, plot=False):
    """
    Smooth an ordered set of r-z coordinates to avoid sharp edges.

    For all points, where the cos of the angle between the point, and the points ± `bounds` are below `cutoff`, the points within the range are replaced by a smoothed set of points.

    r : 1d array
    z : 1d array
    cutoff : float
        cos of angle
    bounds :  integer
        distance of which to smooth
    bounds_find : integer (optional)
        if given, distance in which to check for edges, bounds otherwise

    Return:
    tuple of 1d-arrays
    The smoothed values, if there where outliers, or the original data otherwise.
    """
    if plot:
        import matplotlib.pyplot as plt

        # print(r, z)
        plt.plot(r, z, "o-", label="input")

    bounds_find = bounds_find or bounds
    bounds0 = bounds
    debs = []
    while True:
        outliers = []
        for b in range(1, bounds_find + 1):
            # print(find_outliers(r, z, cutoff, b))
            outliers += find_outliers(r, z, cutoff, b)
        if plot:
            # print(np.array([(r[i], z[i]) for i in outliers]).T)
            if outliers:
                plt.plot(
                    *np.array([(r[i], z[i]) for i in outliers]).T, "x", label="outlier"
                )
        splitedlist = generate_points_tosmooth(bounds, outliers, len(r))
        if plot:
            # plt.plot(
            # print(splitedlist)
            pass
        a, b, fail, deb = gen_newr(r, z, splitedlist, bounds, plot)
        if not fail:
            ret = a, b
            break
        r, z = a, b
        bounds += 1
        debs += [deb[-1]]
        if bounds > bounds0 * 5:
            plt.legend()
            plt.gca().set_aspect("equal")
            plt.title("failed")
            alldeb.append(debs)
            return

    alldeb.append(debs)
    if plot:
        plt.plot(*ret, "o-", label="new")
        plt.legend()
        plt.gca().set_aspect("equal")
        # plt.show()
    return ret


if __name__ == "__main__":
    import sys

    import matplotlib.pyplot as plt

    import zoidberg as zb

    fn = sys.argv[1]
    dat = np.loadtxt(fn)
    dat.shape = (-1, 2, dat.shape[1])
    # line = zb.rzline.RZline(*dat[1])
    j = 0
    for i in range(len(dat)):
        r, z = dat[i]
        if find_outliers(r, z, 0.75, 10):
            plt.figure()
            smooth(r, z, cutoff=0.75, bounds=10, plot=True)
            j += 1
            print(f"figure {j}")
            # if j == 2:
            #    break

    if any([len(x) for x in alldeb]):
        f, axs = plt.subplots(3, 1)
        for deb in alldeb:
            x = np.array(deb)
            print(deb)
            print(x.shape)
            for a, b in zip(axs, x.T):
                a.plot(b)
    plt.show()



from math import gamma

import numpy as np
from sympy import (And, Piecewise, Sum, Symbol, atan2, cos, diff, factorial,
                   lambdify, log, pi, sin, sqrt)

from . import boundary


class MagneticField(object):
    """Represents a magnetic field in either Cartesian or cylindrical
    geometry

    This is the base class, you probably don't want to instantiate one
    of these directly. Instead, create an instance of one of the
    subclasses.

    Functions which can be overridden

    - Bxfunc = Function for magnetic field in x
    - Bzfunc = Function for magnetic field in z
    - Byfunc = Function for magnetic field in y (default = 1.)
    - Rfunc = Function for major radius. If None, y is in meters

    Attributes
    ----------
    boundary
        An object with an "outside" function. See :py:obj:`zoidberg.boundary`

    attributes : A dictionary of string -> function(x,z,phi)
         Contains attributes to be written to the output

    See Also
    --------
    Slab : A straight field in normal Cartesian coordinates
    CurvedSlab : A field in curvilinear coordinates
    StraightStellarator : A rotating ellipse stellarator without curvature
    RotatingEllipse : A rotating ellipse stellarator with curvature
    VMEC : A numerical field from a VMEC equilibrium file
    GEQDSK : A numerical field from an EFIT g-file

    """

    boundary = boundary.NoBoundary()  # An optional Boundary object
    attributes = {}

    def Bxfunc(self, x, z, phi):
        """Magnetic field in x direction at given coordinates

        Parameters
        ----------
        x, z, phi : array_like
            X, Z, and toroidal coordinates

        Returns
        -------
        ndarray
            X-component of the magnetic field

        """
        return np.zeros(x.shape)

    def Byfunc(self, x, z, phi):
        """Magnetic field in y direction at given coordinates

        Parameters
        ----------
        x, z, phi : array_like
            X, Z, and toroidal coordinates

        Returns
        -------
        ndarray
            Y-component of the magnetic field

        """
        return np.ones(x.shape)

    def Bzfunc(self, x, z, phi):
        """Magnetic field in z direction at given coordinates

        Parameters
        ----------
        x, z, phi : array_like
            X, Z, and toroidal coordinates

        Returns
        -------
        ndarray
            Z-component of the magnetic field

        """
        return np.zeros(x.shape)

    def Rfunc(self, x, z, phi):
        """Major radius [meters]

        Returns None if in Cartesian coordinates

        Parameters
        ----------
        x, z, phi : array_like
            X, Z, and toroidal coordinates

        Returns
        -------
        ndarray
            The major radius

        """
        return None

    def pressure(self, x, z, phi):
        """Pressure [Pascals]

        Parameters
        ----------
        x, z, phi : array_like
            X, Z, and toroidal coordinates

        Returns
        -------
        ndarray
            The plasma pressure

        """
        return 0.0

    def Bmag(self, x, z, phi):
        """Magnitude of the magnetic field

        .. math ::

           Bmag = \\sqrt(B_x^2 + B_y^2 + B_z^2)

        Parameters
        ----------
        x, z, phi : array_like
            X, Z, and toroidal coordinates

        Returns
        -------
        ndarray
            The magnitude of the magnetic field

        """
        return np.sqrt(
            self.Bxfunc(x, z, phi) ** 2
            + self.Byfunc(x, z, phi) ** 2
            + self.Bzfunc(x, z, phi) ** 2
        )

    def field_direction(self, pos, ycoord, flatten=False):
        """Calculate the direction of the magnetic field
        Returns the change in x with phi and change in z with phi

        Parameters
        ----------
        pos : ndarray
            2-D NumPy array, with the second dimension being [x,z],
            with x and z in meters
        ycoord : float
            Toroidal angle in radians if cylindrical coordinates,
            metres if Cartesian
        flatten : bool, optional
            If True, return a flattened form of the vector
            components. This is useful for passing to
            :py:obj:`~zoidberg.fieldtracer.FieldTracer`

        Returns
        -------
        (dx/dy, dz/dy) : list of floats or ndarray
            - ``= (R*Bx/Bphi, R*Bz/Bphi)`` if cylindrical
            - ``= (Bx/By, Bz/By)`` if Cartesian

        """

        # Input array must have an even number of points
        assert len(pos) % 2 == 0

        if flatten:
            position = pos.reshape((-1, 2))
            x = position[:, 0]
            z = position[:, 1]
        else:
            x, z = pos

        if hasattr(self, "Bxyzfunc"):
            Bx, By, Bz = self.Bxyzfunc(x, z, ycoord)
        else:
            Bx = self.Bxfunc(x, z, ycoord)
            By = self.Byfunc(x, z, ycoord)
            Bz = self.Bzfunc(x, z, ycoord)

        Rmaj = self.Rfunc(x, z, ycoord)  # Major radius. None if Cartesian
        if Rmaj is None:
            Rmaj = 1

        if np.amin(np.abs(By)) < 1e-8:
            # Very small By
            raise ValueError(
                "Small By ({}) at (x={}, y={}, z={})".format(By, x, ycoord, z)
            )

        R_By = Rmaj / By
        # Rate of change of x location [m] with y angle [radians]
        dxdphi = R_By * Bx
        # Rate of change of z location [m] with y angle [radians]
        dzdphi = R_By * Bz

        if flatten:
            result = np.column_stack((dxdphi, dzdphi)).flatten()
        else:
            result = [dxdphi, dzdphi]

        return result


class Slab(MagneticField):
    """Represents a magnetic field in an infinite flat slab

    Magnetic field in ``z = Bz + (x - xcentre) * Bzprime``

    Coordinates (x,y,z) assumed to be Cartesian, all in metres

    Parameters
    ----------
    By : float, optional
        Magnetic field in y direction
    Bz : float, optional
        Magnetic field in z at xcentre
    xcentre : float, optional
        Reference x coordinate
    Bzprime : float, optional
        Rate of change of Bz with x

    """

    def __init__(self, By=1.0, Bz=0.1, xcentre=0.0, Bzprime=1.0):
        By = float(By)
        Bz = float(Bz)
        xcentre = float(xcentre)
        Bzprime = float(Bzprime)

        self.By = By
        self.Bz = Bz
        self.xcentre = xcentre
        self.Bzprime = Bzprime

    def Bxfunc(self, x, z, phi):
        return np.zeros(x.shape)

    def Byfunc(self, x, z, phi):
        return np.full(x.shape, self.By)

    def Bzfunc(self, x, z, phi):
        return self.Bz + (x - self.xcentre) * self.Bzprime


class CurvedSlab(MagneticField):
    """
    Represents a magnetic field in a curved slab geometry

    Magnetic field in ``z = Bz + (x - xcentre) * Bzprime``

    x  - Distance in radial direction [m]
    y  - Azimuthal (toroidal) angle
    z  - Height [m]

    Parameters
    ----------
    By : float
        Magnetic field in y direction
    Bz : float
        Magnetic field in z at xcentre (float)
    xcentre : float
        Reference x coordinate
    Bzprime : float
        Rate of change of Bz with x
    Rmaj : float
        Major radius of the slab

    """

    def __init__(self, By=1.0, Bz=0.1, xcentre=0.0, Bzprime=1.0, Rmaj=1.0):
        By = float(By)
        Bz = float(Bz)
        xcentre = float(xcentre)
        Bzprime = float(Bzprime)
        Rmaj = float(Rmaj)

        self.By = By
        self.Bz = Bz
        self.xcentre = xcentre
        self.Bzprime = Bzprime
        self.Rmaj = Rmaj

        # Set poloidal magnetic field
        # Bpx = self.Bp + (self.grid.xarray-self.grid.Lx/2.) * self.Bpprime
        # self.Bpxy = np.resize(Bpx, (self.grid.nz, self.grid.ny, self.grid.nx))
        # self.Bpxy = np.transpose(self.Bpxy, (2,1,0))
        # self.Bxy = np.sqrt(self.Bpxy**2 + self.Bt**2)

    def Bxfunc(self, x, z, phi):
        return np.zeros(x.shape)

    def Byfunc(self, x, z, phi):
        return np.full(x.shape, self.By)

    def Bzfunc(self, x, z, phi):
        return self.Bz + (x - self.xcentre) * self.Bzprime

    def Rfunc(self, x, z, phi):
        return np.full(x.shape, self.Rmaj)


class StraightStellarator(MagneticField):
    """A "rotating ellipse" stellarator without curvature

    Parameters
    ----------
    xcentre : float, optional
        Middle of the domain in x [m]
    zcentre : float, optional
        Middle of the domain in z [m]
    radius : float, optional
        Radius of coils [meters]
    yperiod : float, optional
        The period over which the coils return to their original position
    I_coil : float, optional
        Current in each coil

    """

    def coil(self, xcentre, zcentre, radius, angle, iota, I):
        """Defines a single coil

        Parameters
        ----------
        radius : float
            Radius to coil
        angle : float
            Initial angle of coil
        iota : float
            Rotational transform of coil
        I : float
            Current through coil

        Returns
        -------
        (x, z) - x, z coordinates of coils along phi
        """

        return (
            xcentre + radius * cos(angle + iota * self.phi),
            zcentre + radius * sin(angle + iota * self.phi),
            I,
        )

    def __init__(
        self,
        xcentre=0.0,
        zcentre=0.0,
        radius=0.8,
        yperiod=np.pi,
        I_coil=0.05,
        smooth=False,
        smooth_args={},
    ):
        xcentre = float(xcentre)
        zcentre = float(zcentre)
        radius = float(radius)
        yperiod = float(yperiod)

        iota = 2.0 * np.pi / yperiod

        self.x = Symbol("x")
        self.z = Symbol("z")
        self.y = Symbol("y")
        self.r = Symbol("r")
        self.r = (self.x**2 + self.z**2) ** (0.5)
        self.phi = Symbol("phi")

        self.xcentre = xcentre
        self.zcentre = zcentre
        self.radius = radius

        # Four coils equally spaced, alternating direction for current
        self.coil_list = [
            self.coil(
                xcentre,
                zcentre,
                radius,
                n * pi,
                iota,
                ((-1) ** np.mod(i, 2)) * I_coil,
            )
            for i, n in enumerate(np.arange(4) / 2.0)
        ]

        A = 0.0
        Bx = 0.0
        Bz = 0.0

        for c in self.coil_list:
            xc, zc, Ic = c
            r2 = (self.x - xc) ** 2 + (self.z - zc) ** 2
            theta = atan2(self.z - zc, self.x - xc)  # Angle relative to coil

            A -= Ic * 0.1 * log(r2)

            B = Ic * 0.2 / sqrt(r2)

            Bx += B * sin(theta)
            Bz -= B * cos(theta)

        self.Afunc = lambdify((self.x, self.z, self.phi), A, "numpy")

        self.Bxfunc = lambdify((self.x, self.z, self.phi), Bx, "numpy")
        self.Bzfunc = lambdify((self.x, self.z, self.phi), Bz, "numpy")


class RotatingEllipse(MagneticField):
    """A "rotating ellipse" stellarator
    Parameters
    ----------
    xcentre : float, optional
        Middle of the domain in x [m]
    zcentre : float, optional
        Middle of the domain in z [m]
    radius : float, optional
        Radius of coils [meters]
    yperiod : float, optional
        The period over which the coils return to their original position
    I_coil : float, optional
        Current in each coil
    Btor : float, optional
        Toroidal magnetic field strength
    """

    def coil(self, xcentre, zcentre, radius, angle, iota, I):
        """Defines a single coil
        Parameters
        ----------
        radius : float
            Radius to coil
        angle : float
            Initial angle of coil
        iota : float
            Rotational transform of coil
        I : float
            Current through coil
        Returns
        -------
        (x, z) - x, z coordinates of coils along phi
        """

        return (
            xcentre + radius * cos(angle + iota * self.phi),
            zcentre + radius * sin(angle + iota * self.phi),
            I,
        )

    def __init__(
        self,
        xcentre=0.0,
        zcentre=0.0,
        radius=0.8,
        yperiod=np.pi,
        I_coil=0.05,
        Btor=1.0,
        smooth=False,
        smooth_args={},
    ):
        xcentre = float(xcentre)
        zcentre = float(zcentre)
        radius = float(radius)
        yperiod = float(yperiod)
        Btor = float(Btor)

        iota = 2.0 * np.pi / yperiod

        self.x = Symbol("x")
        self.z = Symbol("z")
        self.y = Symbol("y")
        self.r = Symbol("r")
        self.r = (self.x**2 + self.z**2) ** (0.5)
        self.phi = Symbol("phi")

        self.xcentre = xcentre
        self.zcentre = zcentre
        self.radius = radius

        # Four coils equally spaced, alternating direction for current
        self.coil_list = [
            self.coil(
                xcentre,
                zcentre,
                radius,
                n * pi,
                iota,
                ((-1) ** np.mod(i, 2)) * I_coil,
            )
            for i, n in enumerate(np.arange(4) / 2.0)
        ]

        A = 0.0
        Bx = 0.0
        Bz = 0.0

        for c in self.coil_list:
            xc, zc, Ic = c
            rc = (xc**2 + zc**2) ** (0.5)
            r2 = (self.x - xc) ** 2 + (self.z - zc) ** 2
            theta = atan2(self.z - zc, self.x - xc)  # Angle relative to coil

            A -= Ic * 0.1 * log(r2)

            B = Ic * 0.2 / sqrt(r2)

            Bx += B * sin(theta)
            Bz -= B * cos(theta)

        By = Btor / self.x
        self.Afunc = lambdify((self.x, self.z, self.phi), A, "numpy")

        self.Bxfunc = lambdify((self.x, self.z, self.phi), Bx, "numpy")
        self.Bzfunc = lambdify((self.x, self.z, self.phi), Bz, "numpy")
        self.Byfunc = lambdify((self.x, self.z, self.phi), By, "numpy")

    def Rfunc(self, x, z, phi):
        return np.full(x.shape, x)


class DommaschkPotentials(MagneticField):
    """A magnetic field generator using the Dommaschk potentials.
    Parameters
    ----------
    A: Coefficient matrix for the torodial and polidial harmonics. Form: (m,l,(a,b,c,d))
    R_0: major radius [m]
    B_0: magnetic field on axis [T]

    Important Methods
    -----------------
    Bxfunc/Byfunc/Bzfunc(x,z,y):
        Returns magnetic field in radial/torodial/z-direction
    Sfunc(x,z,y):
        Returns approximate magnetic surface invariant for Dommaschk potentials.
        Use this to visualize flux surfaces



    """

    def __init__(self, A, R_0=1.0, B_0=1.0):
        self.R_0 = R_0
        self.B_0 = B_0

        self.R = Symbol("R")
        self.phi = Symbol("phi")
        self.Z = Symbol("Z")

        self.m = Symbol("m")
        self.l = Symbol("l")
        self.n = Symbol("n")
        self.k = Symbol("k")

        self.A = A

        self.P = (
            self.U(self.A)
            .doit()
            .subs([(self.R, self.R / self.R_0), (self.Z, self.Z / self.R_0)])
        )
        self.P_hat = (
            self.U_hat(self.A)
            .doit()
            .subs([(self.R, self.R / self.R_0), (self.Z, self.Z / self.R_0)])
        )

        S = 0.5 * (
            log(self.R / self.R_0) ** 2 + (self.Z / self.R_0) ** 2
        ) - self.R / self.R_0 * (
            log(self.R / self.R_0) * self.R_0 * diff(self.P_hat, self.R)
            + self.Z * self.R / self.R_0 * diff(self.P_hat, self.Z)
        )

        Bx = R_0 * diff(self.P, self.R)
        By = R_0 / self.R * diff(self.P, self.phi)
        Bz = R_0 * diff(self.P, self.Z)

        self.Sf = lambdify((self.R, self.phi, self.Z), S, "numpy")

        self.Bxf = lambdify((self.R, self.phi, self.Z), Bx, "numpy")
        self.Byf = lambdify((self.R, self.phi, self.Z), By, "numpy")
        self.Bzf = lambdify((self.R, self.phi, self.Z), Bz, "numpy")

    def Bxfunc(self, x, z, phi):
        return self.Bxf(x, phi, z) / self.Byf(self.R_0, 0, 0) * self.B_0

    def Byfunc(self, x, z, phi):
        return self.Byf(x, phi, z) / self.Byf(self.R_0, 0, 0) * self.B_0

    def Bzfunc(self, x, z, phi):
        return self.Bzf(x, phi, z) / self.Byf(self.R_0, 0, 0) * self.B_0

    def Sfunc(self, x, z, y):
        """
        Parameters
        ----------
        x: radial coordinates normalized to R_0
        z: binormal coordinate
        y: torodial angle normalized to 2*pi

        Returns
        -------
        Approximate magnetic surface invariant S at location (x,z,y).
        This is from the original Dommaschk paper.
        Use to visualize flux surfaces
        """
        return self.Sf(x, y, z)

    def Rfunc(self, x, z, phi):
        """
        Parameters
        ----------
        x: radial coordinates normalized to R_0
        z: binormal coordinate
        y: torodial angle normalized to 2*pi

        Returns
        -------
        Radial coordinate x
        """

        return x

    def CD(self, m, k):
        """
        Parameters
        ----------
        m: torodial harmonic
        k: summation index in D

        Returns:
        --------
        Sympy function CD_mk (R) (Dirichlet boudary conditions)
        """

        alpha = lambda n, b: (
            (-1.0) ** n / (gamma(b + n + 1) * gamma(n + 1) * 2.0 ** (2 * n + b))
            if (n >= 0)
            else 0.0
        )
        alpha_st = lambda n, b: alpha(n, b) * (2 * n + b)

        beta = lambda n, b: (
            gamma(b - n) / (gamma(n + 1) * 2.0 ** (2 * n - b + 1))
            if (n >= 0 and n < b)
            else 0.0
        )
        beta_st = lambda n, b: beta(n, b) * (2 * n - b)

        delta = lambda n, b: (
            alpha(n, b) * np.sum([1.0 / i + 1.0 / (b + i) for i in range(1, n + 1)]) / 2
            if (n > 0)
            else 0.0
        )
        delta_st = lambda n, b: delta(n, b) * (2 * n + b)

        CD = log(1)
        for j in range(k + 1):
            CD += -(
                alpha(j, m)
                * (
                    alpha_st(k - m - j, m) * log(self.R)
                    + delta_st(k - m - j, m)
                    - alpha(k - m - j, m)
                )
                - delta(j, m) * alpha_st(k - m - j, m)
                + alpha(j, m) * beta_st(k - j, m)
            ) * self.R ** (2 * j + m) + beta(j, m) * alpha_st(k - j, m) * self.R ** (
                2 * j - m
            )

        return CD

    def CN(self, m, k):
        """
        Parameters
        ----------
        m: torodial harmonic
        k: summation index in N

        Returns:
        --------
        Sympy function CN_mk (R) (Neumann boundary conditions)
        """

        alpha = lambda n, b: (
            (-1.0) ** n / (gamma(b + n + 1) * gamma(n + 1) * 2.0 ** (2 * n + b))
            if (n >= 0)
            else 0.0
        )
        alpha_st = lambda n, b: alpha(n, b) * (2 * n + b)

        beta = lambda n, b: (
            gamma(b - n) / (gamma(n + 1) * 2.0 ** (2 * n - b + 1))
            if (n >= 0 and n < b)
            else 0.0
        )
        beta_st = lambda n, b: beta(n, b) * (2 * n - b)

        delta = lambda n, b: (
            alpha(n, b) * np.sum([1.0 / i + 1.0 / (b + i) for i in range(1, n + 1)]) / 2
            if (n > 0)
            else 0.0
        )
        delta_st = lambda n, b: delta(n, b) * (2 * n + b)

        CN = log(1)
        for j in range(k + 1):
            CN += (
                alpha(j, m) * (alpha(k - m - j, m) * log(self.R) + delta(k - m - j, m))
                - delta(j, m) * alpha(k - m - j, m)
                + alpha(j, m) * beta(k - j, m)
            ) * self.R ** (2 * j + m) - beta(j, m) * alpha(k - j, m) * self.R ** (
                2 * j - m
            )

        return CN

    def D(self, m, n):
        """
        Parameters
        ----------
        m: torodial mode number
        n: summation index in  V

        Returns:
        --------
        Sympy function D_mn (R, Z) (Dirichlet boundary conditions)
        """

        D = log(1)
        k_arr = np.arange(0, int(n / 2) + 1, 1)

        for k in k_arr:
            D += (self.Z ** (n - 2 * k)) / factorial(n - 2 * k) * self.CD(m, k)

        return D

    def N(self, m, n):
        """
        Parameters
        ----------
        m: torodial mode number
        n: summation index in V

        Returns:
        --------
        Sympy function N_mn (R, Z) (Neumann boundary conditions)
        """

        N = log(1)
        k_arr = np.arange(0, int(n / 2) + 1, 1)

        for k in k_arr:
            N += (self.Z ** (n - 2 * k)) / factorial(n - 2 * k) * self.CN(m, k)

        return N

    def V(self, m, l, a, b, c, d):
        """
        Parameters
        ----------
        m: torodial mode number
        l: polodial mode number
        a,b,c,d: Coefficients for m,l-th Dommaschk potential (elements of matrix A)

        Returns:
        --------
        Sympy function V_ml
        """

        V = (a * cos(m * self.phi) + b * sin(m * self.phi)) * self.D(m, l) + (
            c * cos(m * self.phi) + d * sin(m * self.phi)
        ) * self.N(m, l - 1)

        return V

    def U(self, A):
        """
        Parameters
        ----------
        A: Coefficient matrix for the torodial and polidial harmonics. Form: (m,l,(a,b,c,d))

        Returns
        -----------------
        U: Superposition of all modes given in A

        """
        U = self.phi
        for i in range(A.shape[0]):
            for j in range(A.shape[1]):
                if A[i, j, 0] or A[i, j, 1] or A[i, j, 2] or A[i, j, 3] != 0:
                    U += self.V(i, j, A[i, j, 0], A[i, j, 1], A[i, j, 2], A[i, j, 3])

        return U

    def V_hat(self, m, l, a, b, c, d):
        """
        Parameters
        ----------
        m: torodial mode number
        l: polodial mode number
        a,b,c,d: Coefficients for m,l-th Dommaschk potential (elements of matrix A)

        Returns:
        --------
        Sympy function V_hat_ml; Similar to V; needed for calculation of magnetic surface invariant S
        """

        V = (
            a * cos(m * self.phi - np.pi / 2) + b * sin(m * self.phi - np.pi / 2)
        ) * self.D(m, l) + (
            c * cos(m * self.phi - np.pi / 2) + d * sin(m * self.phi - np.pi / 2)
        ) * self.N(
            m, l - 1
        )

        return V

    def U_hat(self, A):
        """
        Parameters
        ----------
        A: Coefficient matrix for the torodial and polidial harmonics. Form: (m,l,(a,b,c,d))

        Returns
        -----------------
        U: Superposition of all modes given in A

        """

        U = log(1)
        for i in range(A.shape[0]):
            if i != 0:
                i_inv = 1 / i
            else:
                i_inv = np.nan
            for j in range(A.shape[1]):
                if A[i, j, 0] or A[i, j, 1] or A[i, j, 2] or A[i, j, 3] != 0:
                    U += self.V_hat(
                        i, j, A[i, j, 0], A[i, j, 1], A[i, j, 2], A[i, j, 3]
                    ) * Piecewise((self.phi, i == 0), (i_inv, i > 0))

        return U


class Screwpinch(MagneticField):
    def __init__(self, xcentre=1.5, zcentre=0.0, shear=0, yperiod=2 * np.pi, Btor=1.0):
        self.x = Symbol("x")
        self.z = Symbol("z")
        self.y = Symbol("y")
        self.r = Symbol("r")
        self.r = ((self.x - xcentre) ** 2 + (self.z - zcentre) ** 2) ** (0.5)

        self.phi = Symbol("phi")

        alpha = shear
        self.theta = atan2(self.z - zcentre, self.x - xcentre)
        A = alpha * self.r**2
        Bx = -alpha * self.r * self.r * sin(self.theta)
        Bz = alpha * self.r * self.r * cos(self.theta)
        By = Btor / self.x

        self.Afunc = lambdify((self.x, self.z, self.phi), A, "numpy")
        self.Bxfunc = lambdify((self.x, self.z, self.phi), Bx, "numpy")
        self.Bzfunc = lambdify((self.x, self.z, self.phi), Bz, "numpy")
        self.Byfunc = lambdify((self.x, self.z, self.phi), By, "numpy")

    def Rfunc(self, x, z, phi):
        return np.full(x.shape, x)


class VMEC(MagneticField):
    """A numerical magnetic field from a VMEC equilibrium file

    Parameters
    ----------
    vmec_file : str
        Name of the VMEC file to read
    ntheta : int, optional
        Number of theta points to use (default: use 'mpol' from VMEC file)n
    nzeta : int, optional
        Number of zeta points to use (default: use 'ntor' from VMEC file)
    nr : int
        Number of R points to use
    nz : int
        Number of Z points to use

    """

    def __rolling_average(self, field):
        result = np.zeros_like(field)
        result[:, 0] = field[:, 1] - 0.5 * field[:, 2]
        result[:, 2:-2] = 0.5 * (field[:, 2:-2] + field[:, 3:-1])
        result[:, -1] = 2 * field[:, -2] - field[:, -3]
        return result

    def cfunct(self, field):
        """VMEC DCT"""
        ns = field.shape[0]
        lt = self.theta.size
        lz = self.zeta.size
        # Create mode x angle arrays
        mt = self.xm[:, np.newaxis] * self.theta
        nz = self.xn[:, np.newaxis] * self.zeta
        # Create Trig Arrays
        cosmt = np.cos(mt)
        sinmt = np.sin(mt)
        cosnz = np.cos(nz)
        sinnz = np.sin(nz)
        # Calculate the transform
        f = np.zeros((ns, lt, lz))
        for k, field_slice in enumerate(field):
            rmn = np.repeat(field_slice[:, np.newaxis], lt, axis=1)
            a = rmn * cosmt
            b = np.dot(a.T, cosnz)
            # print("a: {}, b: {}".format(a.shape, b.shape))
            c = rmn * sinmt
            d = np.dot(c.T, sinnz)
            # print("c: {}, d: {}".format(c.shape, d.shape))
            f[k, :, :] = b - d
        return f

    def sfunct(self, field):
        """VMEC DST"""
        ns = field.shape[0]
        lt = self.theta.size
        lz = self.zeta.size
        # Create mode x angle arrays
        mt = self.xm[:, np.newaxis] * self.theta
        nz = self.xn[:, np.newaxis] * self.zeta
        # Create Trig Arrays
        cosmt = np.cos(mt)
        sinmt = np.sin(mt)
        cosnz = np.cos(nz)
        sinnz = np.sin(nz)
        # Calculate the transform
        f = np.zeros((ns, lt, lz))
        for k, field_slice in enumerate(field):
            rmn = np.repeat(field_slice[:, np.newaxis], lt, axis=1)
            a = rmn * sinmt
            b = np.dot(a.T, cosnz)
            c = rmn * cosmt
            d = np.dot(c.T, sinnz)
            f[k, :, :] = b + d
        return f

    def read_vmec_file(self, vmec_file, ntheta=None, nzeta=None):
        """Read a VMEC equilibrium file"""
        from boututils.datafile import DataFile

        # Read necessary stuff
        with DataFile(vmec_file, write=False) as f:
            self.xm = f["xm"].T
            self.xn = f["xn"].T
            ns = int(f["ns"])
            xm_big = np.repeat(self.xm[:, np.newaxis], ns, axis=1)
            xn_big = np.repeat(self.xn[:, np.newaxis], ns, axis=1)
            # s and c seem to swap meanings here...
            rumns = -f["rmnc"].T * xm_big
            rvmns = -f["rmnc"].T * xn_big
            zumnc = f["zmns"].T * xm_big
            zvmnc = f["zmns"].T * xn_big

            try:
                iasym = f["iasym"]
            except KeyError:
                iasym = 0

            if iasym:
                rumnc = -f["rmns"].T * xm_big
                rvmnc = -f["rmns"].T * xn_big
                zumns = f["zmnc"].T * xm_big
                zvmns = f["zmnc"].T * xn_big

            bsupumnc = self.__rolling_average(f["bsupumnc"].T).T
            bsupvmnc = self.__rolling_average(f["bsupvmnc"].T).T

            if ntheta is None:
                self.ntheta = int(f["mpol"])
            else:
                self.ntheta = ntheta

            if nzeta is None:
                self.nzeta = int(f["ntor"]) + 1
            else:
                self.nzeta = nzeta

            self.theta = np.linspace(0, 2 * np.pi, self.ntheta)
            self.zeta = np.linspace(0, 2 * np.pi, self.nzeta)
            # R, Z on (s, theta, zeta)
            self.r_stz = self.cfunct(f["rmnc"])
            self.z_stz = self.sfunct(f["zmns"])

            bu = self.cfunct(bsupumnc)
            bv = self.cfunct(bsupvmnc)

            drdu = self.sfunct(rumns.T)
            drdv = self.sfunct(rvmns.T)
            dzdu = self.cfunct(zumnc.T)
            dzdv = self.cfunct(zvmnc.T)
            if iasym:
                self.r_stz = self.r_stz + self.sfunct(f["rmnc"])
                drdu = drdu + self.cfunct(rumnc.T)
                drdv = drdv + self.cfunct(rvmnc.T)
                self.z_stz = self.z_stz + self.cfunct(f["zmnc"])
                dzdu = dzdu + self.sfunct(zumns.T)
                dzdv = dzdv + self.sfunct(zvmns.T)

        # Convert to bR, bZ, bphi
        self.br = bu * drdu + bv * drdv
        self.bphi = self.r_stz * bv
        self.bz = bu * dzdu + bv * dzdv

    def __init__(self, vmec_file, ntheta=None, nzeta=None, nr=32, nz=32):
        # Only needed here
        from scipy.interpolate import RegularGridInterpolator, griddata

        self.read_vmec_file(vmec_file, ntheta, nzeta)

        self.nr = nr
        self.nz = nz

        # Make a new rectangular grid in (R,Z)
        self.r_1D = np.linspace(self.r_stz.min(), self.r_stz.max(), nr)
        self.z_1D = np.linspace(self.z_stz.min(), self.z_stz.max(), nz)
        self.R_2D, self.Z_2D = np.meshgrid(self.r_1D, self.z_1D, indexing="ij")

        # First, interpolate the magnetic field components onto (R,Z)
        self.br_rz = np.zeros((nr, nz, self.nzeta))
        self.bz_rz = np.zeros((nr, nz, self.nzeta))
        self.bphi_rz = np.zeros((nr, nz, self.nzeta))
        # No need to interpolate in zeta, so do this one slice at a time
        for k, (br, bz, bphi, r, z) in enumerate(
            zip(self.br.T, self.bz.T, self.bphi.T, self.r_stz.T, self.z_stz.T)
        ):
            points = np.column_stack((r.flatten(), z.flatten()))
            self.br_rz[..., k] = griddata(
                points,
                br.flatten(),
                (self.R_2D, self.Z_2D),
                method="linear",
                fill_value=0.0,
            )
            self.bz_rz[..., k] = griddata(
                points,
                bz.flatten(),
                (self.R_2D, self.Z_2D),
                method="linear",
                fill_value=0.0,
            )
            self.bphi_rz[..., k] = griddata(
                points,
                bphi.flatten(),
                (self.R_2D, self.Z_2D),
                method="linear",
                fill_value=1.0,
            )

        # Now we have a regular grid in (R,Z,phi) (as zeta==phi), so
        # we can get an interpolation function in 3D
        points = (self.r_1D, self.z_1D, self.zeta)

        self.br_interp = RegularGridInterpolator(
            points, self.br_rz, bounds_error=False, fill_value=0.0
        )
        self.bz_interp = RegularGridInterpolator(
            points, self.bz_rz, bounds_error=False, fill_value=0.0
        )
        self.bphi_interp = RegularGridInterpolator(
            points, self.bphi_rz, bounds_error=False, fill_value=1.0
        )

    def Bxfunc(self, x, z, phi):
        phi = np.mod(phi, 2.0 * np.pi)
        return self.br_interp((x, z, phi))

    def Bzfunc(self, x, z, phi):
        phi = np.mod(phi, 2.0 * np.pi)
        return self.bz_interp((x, z, phi))

    def Byfunc(self, x, z, phi):
        phi = np.mod(phi, 2.0 * np.pi)
        return self.bphi_interp((x, z, phi))

    def Rfunc(self, x, z, phi):
        """
        Major radius
        """
        return x


class SmoothedMagneticField(MagneticField):
    """Represents a magnetic field which is smoothed so it never leaves
    the boundaries of a given grid.

    Parameters
    ----------
    field : :py:obj:`zoidberg.field.MagneticField`
        A MagneticField object
    grid : :py:obj:`zoidberg.grid.Grid`
        A Grid object
    xboundary : int, optional
        Number of grid points in x over which the magnetic field is smoothed
    zboundary : int, optional
        Number of grid points in x over which the magnetic field is smoothed

    """

    def __init__(self, field, grid, xboundary=None, zboundary=None):
        """ """

        self.field = field
        self.grid = grid
        self.xboundary = xboundary
        self.zboundary = zboundary

    def Bxfunc(self, x, z, phi):
        # Get closest y (phi) grid index
        ind = np.argmin(np.abs(phi - self.grid.ycoords))
        if phi < self.grid.ycoords[ind]:
            ind -= 1
        # phi now between ind and ind+1
        grid_d, y_d = self.grid.getPoloidalGrid(ind)
        grid_u, y_u = self.grid.getPoloidalGrid(ind + 1)

        # Get x,z indices from poloidal grids
        x_d, z_d = grid_d.findIndex(x, z)
        x_u, z_u = grid_d.findIndex(x, z)

    def Byfunc(self, x, z, phi):
        """
        Not modified by smoothing
        """
        return self.field.Byfunc(x, z, phi)

    def Bxfunc(self, x, z, phi):
        pass

    def Rfunc(self, x, z, phi):
        return self.field.Rfunc(x, z, phi)

    def smooth_field_line(self, xa, za):
        """Linearly damp the field to be parallel to the edges of the box

        Should take some parameters to adjust rate of smoothing, etc.
        """

        x_left = (xa - self.xl_inner) / (self.xl_inner - self.xl_outer) + 1
        x_right = (xa - self.xr_inner) / (self.xr_inner - self.xr_outer) + 1
        z_top = (za - self.zt_inner) / (self.zt_inner - self.zt_outer) + 1
        z_bottom = (za - self.zb_inner) / (self.zb_inner - self.zb_outer) + 1

        if xa < self.xl_inner:
            if za < self.zb_inner:
                P = np.min([x_left, z_bottom])
            elif za >= self.zt_inner:
                P = np.min([x_left, z_top])
            else:
                P = x_left
        elif xa >= self.xr_inner:
            if za < self.zb_inner:
                P = np.min([x_right, z_bottom])
            elif za >= self.zt_inner:
                P = np.min([x_right, z_top])
            else:
                P = x_right

        elif za < self.zb_inner:
            P = z_bottom
        elif za > self.zt_inner:
            P = z_top
        else:
            P = 1.0
        if P < 0.0:
            P = 0.0
        return P


class GEQDSK(MagneticField):
    """Read a EFIT G-Eqdsk file for a toroidal equilibrium

    This generates a grid in cylindrical geometry

    Parameters
    ----------
    gfile : str
        Name of the file to open

    """

    def __init__(self, gfile):
        from freeqdsk import geqdsk

        with open(gfile) as f:
            data = geqdsk.read(f)

        # Get the range of major radius
        self.rmin = data["rleft"]
        self.rmax = data["rdim"] + self.rmin

        # Range of height
        self.zmin = data["zmid"] - 0.5 * data["zdim"]
        self.zmax = data["zmid"] + 0.5 * data["zdim"]

        print("Major radius: {0} -> {1} m".format(self.rmin, self.rmax))
        print("Height: {0} -> {1} m".format(self.zmin, self.zmax))

        # Poloidal flux
        self.psi = np.transpose(data["psi"])
        nr, nz = self.psi.shape

        # Normalising factors: psi on axis and boundary
        self.psi_axis = data["simagx"]
        self.psi_bndry = data["sibdry"]

        # Current flux function f = R * Bt
        self.fpol = data["fpol"]

        # Pressure [Pascals]
        self.p = data["pres"]

        self.r = np.linspace(self.rmin, self.rmax, nr)
        self.z = np.linspace(self.zmin, self.zmax, nz)

        # Create a 2D spline interpolation for psi
        from scipy import interpolate

        self.psi_func = interpolate.RectBivariateSpline(self.r, self.z, self.psi)

        # Add to the attributes so that it can be written to file
        self.attributes["psi"] = lambda x, z, phi: self.psi_func(x, z, grid=False)

        # Create a normalised psi array

        self.psinorm = (self.psi - self.psi_axis) / (self.psi_bndry - self.psi_axis)

        # Need to mark areas outside the core as psinorm = 1
        # eg. around coils or in the private flux region
        # Create a boundary

        rb = data["rbdry"]
        zb = data["zbdry"]
        core_bndry = boundary.PolygonBoundaryXZ(rb, zb)

        # Get the points outside the boundary
        rxz, zxz = np.meshgrid(self.r, self.z, indexing="ij")
        outside = core_bndry.outside(rxz, 0.0, zxz)
        self.psinorm[outside] = 1.0

        self.psinorm_func = interpolate.RectBivariateSpline(
            self.r, self.z, self.psinorm
        )

        # Spline for interpolation of f = R*Bt
        psinorm = np.linspace(0.0, 1.0, nr)
        self.f_spl = interpolate.InterpolatedUnivariateSpline(psinorm, self.fpol, ext=3)
        # ext=3 specifies that boundary values are used outside range

        # Spline for interpolation of pressure
        self.p_spl = interpolate.InterpolatedUnivariateSpline(psinorm, self.p, ext=3)

        # Set boundary
        rlim = data["rlim"]
        zlim = data["zlim"]
        if len(rlim) > 0:
            # Create a boundary in X-Z with a polygon representation
            self.boundary = boundary.PolygonBoundaryXZ(rlim, zlim)

    def Bxfunc(self, x, z, phi):
        return -self.psi_func(x, z, dy=1, grid=False) / x

    def Bzfunc(self, x, z, phi):
        return self.psi_func(x, z, dx=1, grid=False) / x

    def Byfunc(self, x, z, phi):
        # Interpolate to get flux surface normalised psi
        psinorm = self.psinorm_func(x, z, grid=False)

        # Interpolate fpol array at values of normalised psi
        if hasattr(psinorm, "shape"):
            return np.reshape(self.f_spl(np.ravel(psinorm)), psinorm.shape) / x

        return self.f_spl(psinorm) / x  # f = R*Bt

    def Rfunc(self, x, z, phi):
        return x

    def pressure(self, x, z, phi):
        # Interpolate to get flux surface normalised psi
        psinorm = self.psinorm_func(x, z, grid=False)

        if hasattr(psinorm, "shape"):
            return np.reshape(self.p_spl(np.ravel(psinorm)), psinorm.shape)

        return self.p_spl(psinorm)


tracer = None


class W7X_vacuum(MagneticField):
    def __init__(
        self,
        nx=128,
        ny=32,
        nz=128,
        x_range=(4.05, 6.55),
        z_range=(-1.35, 1, 35),
        phimax=2.0 * np.pi,
        configuration=0,
        plot_poincare=False,
        include_plasma_field=False,
        wout_file="wout_w7x.0972_0926_0880_0852_+0000_+0000.01.00jh.nc",
    ):
        """
        Get the field for W7X from the webservices.

        Parameters
        ----------
        configuration : int
            The id's are listed here:
            http://webservices.ipp-hgw.mpg.de/docs/fieldlinetracer.html#MagneticConfig
            While the description are at:
            http://svvmec1.ipp-hgw.mpg.de:8080/vmecrest/v1/Coil_currents_1_AA_T_0011.pdf
        """
        import numpy as np
        from scipy.interpolate import RegularGridInterpolator

        ## create 1D arrays of cylindrical coordinates
        r = np.linspace(x_range[0], x_range[-1], nx)
        phi = np.linspace(0, phimax, ny)
        z = np.linspace(z_range[0], z_range[-1], nz)

        ## make those 1D arrays 3D
        rarray, yarray, zarray = np.meshgrid(r, phi, z, indexing="ij")

        ## call vacuum field values
        b_vac = self.field_values(rarray, yarray, zarray, configuration, plot_poincare)
        Bx_vac = b_vac[0]
        By_vac = b_vac[1]
        Bz_vac = b_vac[2]

        if include_plasma_field:
            b_plasma = W7X_vacuum.plasma_field(
                rarray, yarray, zarray, wout_file=wout_file
            )
            Bx_plasma = b_plasma[0]
            By_plasma = b_plasma[1]
            Bz_plasma = b_plasma[2]
        else:
            Bx_plasma = 0
            By_plasma = 0
            Bz_plasma = 0

        Bx = Bx_vac + Bx_plasma
        By = By_vac + By_plasma
        Bz = Bz_vac + Bz_plasma

        # Now we have a field and regular grid in (R,Z,phi) so
        # we can get an interpolation function in 3D
        points = (r, phi, z)

        self.br_interp = RegularGridInterpolator(
            points, Bx, bounds_error=False, fill_value=0.0
        )
        self.bz_interp = RegularGridInterpolator(
            points, Bz, bounds_error=False, fill_value=0.0
        )
        self.bphi_interp = RegularGridInterpolator(
            points, By, bounds_error=False, fill_value=1.0
        )

        # if you want non-interpolated, 3D arrays, make this your return function:
        # return Bx,By,Bz

        # return points, br_interp, bphi_interp, bz_interp

    @classmethod
    def field_values(cls, r, phi, z, configuration=0, plot_poincare=False):
        """This uses the webservices field line tracer to get the vacuum
        magnetic field given 3d arrrays for R, phi, and Z. Only works
        on IPP network

        http://webservices.ipp-hgw.mpg.de/docs/fieldlinetracer.html

        Contact brendan.shanahan@ipp.mpg.de for questions

        """
        import os.path
        import pickle

        import matplotlib.pyplot as plt
        import xarray as xr
        from osa import Client

        global tracer
        tracer = tracer or Client(
            "http://esb.ipp-hgw.mpg.de:8280/services/FieldLineProxy?wsdl"
        )

        nx = r.shape[0]
        ny = phi.shape[1]
        nz = z.shape[2]

        # create (standardized) file name for saving/loading magnetic field.
        fname_old = (
            "B.w7x.{}.{}.{}.{:.2f}-{:.2f}.{:.2f}-{:.2f}.{:.2f}-{:.2f}.dat".format(
                nx,
                ny,
                nz,
                r[0, 0, 0],
                r[-1, 0, 0],
                phi[0, 0, 0],
                phi[0, -1, 0],
                z[0, 0, 0],
                z[0, 0, -1],
            )
        )

        fname = "B.w7x.{}.{}.{}.{}.{:.2f}-{:.2f}.{:.2f}-{:.2f}.{:.2f}-{:.2f}.nc".format(
            configuration,
            nx,
            ny,
            nz,
            r[0, 0, 0],
            r[-1, 0, 0],
            phi[0, 0, 0],
            phi[0, -1, 0],
            z[0, 0, 0],
            z[0, 0, -1],
        )

        if os.path.isfile(fname):
            print("Saved field found, loading from: ", fname)
            with xr.open_dataset(fname) as ds:
                Br = ds["Br"].values
                Bphi = ds["Bphi"].values
                Bz = ds["Bz"].values

        elif os.path.isfile(fname_old):
            print("Saved field found, loading from: ", fname_old)
            with open(fname_old, "rb") as f:
                Br, Bphi, Bz = pickle.load(f)
            with xr.Dataset() as ds:
                ds["Br"] = ("x", "y", "z"), Br
                ds["Bz"] = ("x", "y", "z"), Bz
                ds["Bphi"] = ("x", "y", "z"), Bphi
                ds.to_netcdf(fname)

        else:
            print(
                "No saved field found -- (re)calculating (must be on IPP network for this to work...)"
            )
            print(
                f"Calculating field for Wendelstein 7-X; nx = {nx}, ny = {ny}, nz = {nz}"
            )

            B = cls._calc_B(r, z, phi, configuration)
            Bx, By, Bz = B

            ## Convert to cylindrical coordinates
            Br = Bx * np.cos(phi) + By * np.sin(phi)
            Bphi = -Bx * np.sin(phi) + By * np.cos(phi)

            del Bx
            del By

            ## Save so we don't have to do this every time.
            with xr.Dataset() as ds:
                ds["Br"] = ("x", "y", "z"), Br
                ds["Bz"] = ("x", "y", "z"), Bz
                ds["Bphi"] = ("x", "y", "z"), Bphi
                ds.to_netcdf(fname)

        if plot_poincare:
            ## Poincare plot as done on the web services
            ## Independent of the previously-made field.

            print("Making poincare plot (only works on IPP network)...")
            ## Create configuration objects
            config = tracer.types.MagneticConfig()
            config = _set_config(config, configuration)
            pos = tracer.types.Points3D()

            pos.x1 = np.linspace(5.6, 6.2, 80)
            pos.x2 = np.zeros(80)
            pos.x3 = np.zeros(80)

            poincare = tracer.types.PoincareInPhiPlane()
            poincare.numPoints = 200
            poincare.phi0 = [
                0.0
            ]  ## This is where the poincare plane is (bean=0, triangle = pi/5.)

            task = tracer.types.Task()
            task.step = 0.2
            task.poincare = poincare

            res = tracer.service.trace(pos, config, task, None, None)

            for i in range(0, len(res.surfs)):
                plt.scatter(
                    res.surfs[i].points.x1, res.surfs[i].points.x3, color="black", s=0.1
                )
                plt.show()

        return Br, Bphi, Bz

    def plasma_field(r, phi, z, wout_file="wout.nc"):
        """This uses EXTENDER via the IPP webservices to get the magnetic
        field from the plasma given 3d arrrays for R, phi, and Z. Only
        works on IPP network

        http://webservices.ipp-hgw.mpg.de/docs/extender.html

        Contact brendan.shanahan@ipp.mpg.de for questions

        """
        import os.path
        import pickle

        from osa import Client

        cl = Client("http://esb.ipp-hgw.mpg.de:8280/services/Extender?wsdl")

        vmecURL = "http://svvmec1.ipp-hgw.mpg.de:8080/vmecrest/v1/w7x_ref_1/wout.nc"

        nx = r.shape[0]
        ny = phi.shape[1]
        nz = z.shape[2]

        ### create (standardized) file name for saving/loading magnetic field.
        fname = (
            "B.w7x_plasma_field."
            + str(nx)
            + "."
            + str(ny)
            + "."
            + str(nz)
            + "."
            + "{:.2f}".format(r[0, 0, 0])
            + "-"
            + "{:.2f}".format(r[-1, 0, 0])
            + "."
            + "{:.2f}".format(phi[0, 0, 0])
            + "-"
            + "{:.2f}".format(phi[0, -1, 0])
            + "."
            + "{:.2f}".format(z[0, 0, 0])
            + "-"
            + "{:.2f}".format(z[0, 0, -1])
            + ".dat"
        )

        if os.path.isfile(fname):
            print("Saved field found, loading from: ", fname)
            with open(fname, "rb") as f:
                Br, Bphi, Bz = pickle.load(f)
        else:
            print(
                "No saved plasma field found -- (re)calculating (must be on IPP network for this to work...)"
            )
            print(
                "Calculating plasma field for Wendelstein 7-X; nx = ",
                nx,
                " ny = ",
                ny,
                " nz = ",
                nz,
            )
            print(
                "This part takes AGES... estimate: ",
                nx * ny * nz / 52380.0,
                " minutes.",
            )

            points = cl.types.Points3D()

            ## Extender uses cylindrical coordinates, no need to convert, just flatten.
            points.x1 = np.ndarray.flatten(r)  # x in Cylindrical
            points.x2 = np.ndarray.flatten(phi)  # y in Cylindrical
            points.x3 = np.ndarray.flatten(z)  # z in Cylindrical

            ## call EXTENDER on web services
            # if not (os.path.isfile(wout_file)):
            plasmafield = cl.service.getPlasmaField(None, vmecURL, points, None)
            # else:
            # plasmafield = cl.service.getPlasmaField(wout, None, points, None)

            ## Reshape to 3d array
            Br = np.ndarray.reshape(np.asarray(plasmafield.x1), (nx, ny, nz))
            Bphi = np.ndarray.reshape(np.asarray(plasmafield.x2), (nx, ny, nz))
            Bz = np.ndarray.reshape(np.asarray(plasmafield.x3), (nx, ny, nz))

            ## Save so we don't have to do this every time.
            with open(fname, "wb") as f:
                pickle.dump([Br, Bphi, Bz], f)

        return Br, Bphi, Bz

    def magnetic_axis(self, phi_axis=0, configuration=0):
        from osa import Client

        global tracer
        tracer = tracer or Client(
            "http://esb.ipp-hgw.mpg.de:8280/services/FieldLineProxy?wsdl"
        )

        config = tracer.types.MagneticConfig()
        config = _set_config(config, configuration)
        settings = tracer.types.AxisSettings()
        res = tracer.service.findAxis(0.05, config, settings)

        magnetic_axis_x = np.asarray(res.axis.vertices.x1)  # (m)
        magnetic_axis_y = np.asarray(
            res.axis.vertices.x2
        )  # (m) -- REAL SPACE from an arbitrary start point
        magnetic_axis_z = np.asarray(res.axis.vertices.x3)  # (m)
        magnetic_axis_rmaj = np.sqrt(
            magnetic_axis_x**2 + magnetic_axis_y**2 + magnetic_axis_z**2
        )

        magnetic_axis_r = np.sqrt(
            np.asarray(magnetic_axis_x) ** 2 + np.asarray(magnetic_axis_y**2)
        )
        magnetic_axis_phi = np.arctan(magnetic_axis_y / magnetic_axis_x)

        index = np.where(
            (magnetic_axis_phi >= 0.97 * phi_axis)
            & (magnetic_axis_phi <= 1.03 * phi_axis)
        )
        index = index[0]

        return np.asarray([magnetic_axis_r[index], magnetic_axis_z[index]])[:, 0]

    def Bxfunc(self, x, z, phi):
        phi = np.mod(phi, 2.0 * np.pi)
        return self.br_interp((x, phi, z))

    def Bzfunc(self, x, z, phi):
        phi = np.mod(phi, 2.0 * np.pi)
        return self.bz_interp((x, phi, z))

    def Byfunc(self, x, z, phi):
        phi = np.mod(phi, 2.0 * np.pi)
        # Interpolate to get flux surface normalised psi
        return self.bphi_interp((x, phi, z))

    def Rfunc(self, x, z, phi):
        phi = np.mod(phi, 2.0 * np.pi)
        return x

    @classmethod
    def _calc_B(cls, r, z, phi, configuration):
        x123 = r * np.cos(phi), r * np.sin(phi), z
        x123 = np.array(x123)
        return cls._calc_chunk(x123, configuration)

    @classmethod
    def _calc_chunk(cls, x, configuration):
        from time import sleep

        from osa import Client

        global tracer
        tracer = tracer or Client(
            "http://esb.ipp-hgw.mpg.de:8280/services/FieldLineProxy?wsdl",
            osa_timeout=0.1,
        )

        assert x.shape[0] == 3
        pos = tracer.types.Points3D()
        pos.x1 = x[0].flatten()
        pos.x2 = x[1].flatten()
        pos.x3 = x[2].flatten()

        ## Create configuration objects
        config = tracer.types.MagneticConfig()
        config = _set_config(config, configuration)

        ## Call tracer service
        redo = 2
        while redo:
            try:
                res = tracer.service.magneticField(
                    pos, config, osa_timeout=(10 + len(pos.x1) / 1000)
                )
            except:
                # Catch any error. Different errors might be
                # reported, but we want to retry anyway.
                # Do not except Exception, as that would also
                # ignore control-C, which we do not want to
                # ignore.
                redo -= 1
                if redo == 0:
                    raise
                sleep(0.6 - 0.4 * redo)
            else:
                break
        return np.array((res.field.x1, res.field.x2, res.field.x3)).reshape(x.shape)


def _set_config(config, configuration):
    if isinstance(configuration, int):
        if configuration >= 0:
            config.configIds = configuration
            return config
    known = {
        -1: [1, 1, 1, 1, 1, 1, 1],
        "FMM002": [13423, 13423, 13423, 13423, 13423, -3544, -3544],
    }
    currents = known[configuration]
    currents = np.array([108] * 5 + [36] * 2) * np.array(currents)
    print(currents)
    coils = (
        [160, 165, 170, 175, 180, 185, 190, 195, 200, 205]
        + [161, 166, 171, 176, 181, 186, 191, 196, 201, 206]
        + [162, 167, 172, 177, 182, 187, 192, 197, 202, 207]
        + [163, 168, 173, 178, 183, 188, 193, 198, 203, 208]
        + [164, 169, 174, 179, 184, 189, 194, 199, 204, 209]
        + [210, 212, 214, 216, 218, 220, 222, 224, 226, 228]
        + [211, 213, 215, 217, 219, 221, 223, 225, 227, 229]
        + [230, 231, 232, 233, 234, 235, 236, 237, 238, 239]
        + [350, 241, 351, 352, 353]
    )

    if len(currents) == 7:
        currents = np.array([[x] * 10 for x in currents]).flatten()
    elif len(currents) == 7 + 10 + 5:
        currents = np.append(
            np.array([[x] * 10 for x in currents[:7]]).flatten(), currents[7:]
        )
    elif len(currents) == 10:
        currents = np.append(
            np.array([[x] * 10 for x in currents[:7]]).flatten(),
            np.append(np.array([currents[7:9]] * 5), [currents[9]] * 5),
        )
    if len(currents) == 70:
        coils = coils[:70]
    # I1, I1, I1, I1, I1, I1, I1, I1, I1, I1,
    # I2, I2, I2, I2, I2, I2, I2, I2, I2, I2,
    # I3, I3, I3, I3, I3, I3, I3, I3, I3, I3,
    # I4, I4, I4, I4, I4, I4, I4, I4, I4, I4,
    # I5, I5, I5, I5, I5, I5, I5, I5, I5, I5,
    # IA, IA, IA, IA, IA, IA, IA, IA, IA, IA,
    # IB, IB, IB, IB, IB, IB, IB, IB, IB, IB,
    # Icc1, Icc2, Icc3, Icc4, Icc5, Icc6, Icc7, Icc8, Icc9, Icc10,
    # Itc1, Itc2, Itc3, Itc4, Itc5]
    config.coilsIds = coils
    config.coilsIdsCurrents = np.array(currents)
    return config


class W7X_vacuum_on_demand:
    def __init__(self, configuration):
        self.configuration = configuration
        self.boundary = boundary.NoBoundary()  # An optional Boundary object
        self.attributes = {}

    def Rfunc(self, x, z, phi):
        return x

    def Bmag(self, x, z, phi):
        return np.sqrt(np.sum(self.getB(x, z, phi) ** 2, axis=0))

    def pressure(self, x, z, phi):
        return np.zeros_like(x)

    def getB(self, *pos):
        return W7X_vacuum._calc_B(*pos, self.configuration)

    def Bxfunc(self, x, z, phi):
        B = self.getB(x, z, phi)
        return B[0] * np.cos(phi) + B[1] * np.sin(phi)

    def Byfunc(self, x, z, phi):
        B = self.getB(x, z, phi)
        return -B[0] * np.sin(phi) + B[1] * np.cos(phi)

    def Bxyzfunc(self, x, z, phi):
        B = self.getB(x, z, phi)
        Bx = B[0] * np.cos(phi) + B[1] * np.sin(phi)
        By = -B[0] * np.sin(phi) + B[1] * np.cos(phi)
        return Bx, By, B[2]

    def Bzfunc(self, *pos):
        return self.getB(*pos)[2]


class W7X_VMEC(MagneticField):
    def __init__(
        self,
        nx=512,
        ny=32,
        nz=512,
        x_range=[4.05, 6.55],
        z_range=[-1.35, 1, 35],
        phi_range=[0, 2 * np.pi],
        vmec_id="w7x_ref_171",
    ):
        import numpy as np
        from scipy.interpolate import RegularGridInterpolator

        ## create 1D arrays of cylindrical coordinates
        r = np.linspace(x_range[0], x_range[-1], nx)
        phi = np.linspace(phi_range[0], phi_range[-1], ny)
        z = np.linspace(z_range[0], z_range[-1], nz)

        ## make those 1D arrays 3D
        rarray, yarray, zarray = np.meshgrid(r, phi, z, indexing="ij")

        ## call vacuum field values
        b_vmec = self.field_values(rarray, yarray, zarray, vmec_id)
        Bx_vmec = b_vmec[0]
        By_vmec = b_vmec[1]
        Bz_vmec = b_vmec[2]

        # Now we have a field and regular grid in (R,Z,phi) so
        # we can get an interpolation function in 3D
        points = (r, phi, z)

        self.br_interp = RegularGridInterpolator(
            points, Bx_vmec, bounds_error=False, fill_value=0.0
        )
        self.bz_interp = RegularGridInterpolator(
            points, Bz_vmec, bounds_error=False, fill_value=0.0
        )
        self.bphi_interp = RegularGridInterpolator(
            points, By_vmec, bounds_error=False, fill_value=1.0
        )

    def field_values(self, r, phi, z, vmec_id="w7x_ref_171"):
        from osa import Client

        vmec = Client("http://esb.ipp-hgw.mpg.de:8280/services/vmec_v5?wsdl")

        pos = vmec.types.Points3D()

        # x in Cartesian (real-space)
        pos.x1 = np.ndarray.flatten(
            np.ones((self.nx, self.ny, self.nz)) * r * np.cos(phi)
        )
        # y in Cartesian (real-space)
        pos.x2 = np.ndarray.flatten(
            np.ones((self.nx, self.ny, self.nz)) * r * np.sin(phi)
        )
        pos.x3 = np.ndarray.flatten(z)  # z in Cartesian (real-space)
        b = vmec.service.magneticField(str(vmec_id), pos)

        ## Reshape to 3d array
        Bx = np.ndarray.reshape(np.asarray(b.field.x1), (self.nx, self.ny, self.nz))
        By = np.ndarray.reshape(np.asarray(b.field.x2), (self.nx, self.ny, self.nz))
        Bz = np.ndarray.reshape(np.asarray(b.field.x3), (self.nx, self.ny, self.nz))

        ## Convert to cylindrical coordinates
        Br = Bx * np.cos(phi) + By * np.sin(phi)
        Bphi = -Bx * np.sin(phi) + By * np.cos(phi)

        return Br, Bphi, Bz

    def Bxfunc(self, x, z, phi):
        phi = np.mod(phi, 2.0 * np.pi)
        return self.br_interp((x, phi, z))

    def Bzfunc(self, x, z, phi):
        phi = np.mod(phi, 2.0 * np.pi)
        return self.bz_interp((x, phi, z))

    def Byfunc(self, x, z, phi):
        phi = np.mod(phi, 2.0 * np.pi)
        # Interpolate to get flux surface normalised psi
        return self.bphi_interp((x, phi, z))

    def Rfunc(self, x, z, phi):
        phi = np.mod(phi, 2.0 * np.pi)
        return x
